#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.personal_data.orders.trailing_profiles.trailing_profile as trailing_profile
import octobot_trading.personal_data.orders.trailing_profiles.trailing_profile_types as trailing_profile_types
import octobot_trading.personal_data.orders.trailing_profiles.trailing_price_step as trailing_price_step
import octobot_trading.personal_data.orders.trailing_profiles.filled_take_profit_trailing_profile as \
    filled_take_profit_trailing_profile


def create_trailing_profile(
    profile_type: trailing_profile_types.TrailingProfileTypes,
    profile: dict
) -> trailing_profile.TrailingProfile:
    if profile_type == filled_take_profit_trailing_profile.FilledTakeProfitTrailingProfile.get_type():
        return filled_take_profit_trailing_profile.FilledTakeProfitTrailingProfile.from_dict(profile)
    raise NotImplementedError(f"Unsupported profile_type: {profile_type}")


def create_filled_take_profit_trailing_profile(
    entry_price: decimal.Decimal, take_profit_orders: list
) -> filled_take_profit_trailing_profile.FilledTakeProfitTrailingProfile:
    if not take_profit_orders:
        raise ValueError(f"take_profit_orders can't be empty")
    previous_order = take_profit_orders[0]
    trigger_above = previous_order.trigger_above
    steps = [
        trailing_price_step.TrailingPriceStep(
            float(entry_price), float(previous_order.origin_price), trigger_above
        )
    ]
    for order in take_profit_orders[1:]:
        if order.trigger_above != trigger_above:
            raise ValueError(
                f"trigger_above can't be different from first given order trigger_above value: {trigger_above}"
            )
        steps.append(
            trailing_price_step.TrailingPriceStep(
                # trailing price is the origin price of the previous order
                float(previous_order.origin_price), float(order.origin_price), order.trigger_above
            )
        )
        previous_order = order
    return filled_take_profit_trailing_profile.FilledTakeProfitTrailingProfile(steps)


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import mock
import pytest

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_commons.asyncio_tools as asyncio_tools
from tests.personal_data import DEFAULT_ORDER_SYMBOL
from tests.personal_data.orders.groups import order_mock
from tests.personal_data.orders import created_order
from tests.exchanges import simulated_trader, simulated_exchange_manager
import tests.test_utils.random_numbers as random_numbers


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def oco_group():
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.OneCancelsTheOtherOrderGroup("name",  orders_manager)


async def test_on_fill(oco_group):
    order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order)
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order, ignored_orders=["hello"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    other_order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order,
                                                                                                     other_order])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order)
        order.trader.cancel_order.assert_not_called()
        other_order.trader.cancel_order.assert_called_once_with(
            other_order, ignored_order=order, wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        get_order_from_group_mock.assert_called_once_with(oco_group.name)


async def test_on_fill_no_mock(simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    oco_group = personal_data.OneCancelsTheOtherOrderGroup("name",
                                                           exchange_manager.exchange_personal_data.orders_manager)
    stop_loss = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss.update(
        price=decimal.Decimal(10),
        quantity=decimal.Decimal(1),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=oco_group,
    )
    sell_limit = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit.update(
        price=decimal.Decimal(20),
        quantity=decimal.Decimal(1),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=oco_group,
    )

    for order in [stop_loss, sell_limit]:
        order.exchange_manager.is_backtesting = True  # force update_order_status
        await order.initialize()
        await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(
            order
        )
    price_events_manager = stop_loss.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    # stop loss sell order triggers when price is bellow or equal to its trigger price
    price_events_manager.handle_recent_trades(
        [random_numbers.decimal_random_recent_trade(price=decimal.Decimal(9), timestamp=stop_loss.timestamp)]
    )
    # simulate a real-trader order (that is self-managed therefore an exchange update is not required)
    stop_loss.simulated = False

    # fill stop loss and cancel sell limit
    await asyncio_tools.wait_asyncio_next_cycle()
    assert stop_loss.is_filled()
    assert sell_limit.is_cancelled()


async def test_on_cancel(oco_group):
    order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order)
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order, ignored_orders=["hi"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
        get_order_from_group_mock.reset_mock()
        with pytest.raises(errors.OrderGroupTriggerArgumentError):
            await oco_group.on_cancel(order, ignored_orders=["hi", "ho"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_not_called()
    other_order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order,
                                                                                                     other_order])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order, ignored_orders=["hi"])
        order.trader.cancel_order.assert_not_called()
        other_order.trader.cancel_order.assert_called_once_with(
            other_order, ignored_order="hi", wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        get_order_from_group_mock.assert_called_once_with(oco_group.name)


async def test_oco_adapt_before_order_becoming_active(simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    trader_instance.allow_artificial_orders = False
    trader_instance.enable_inactive_orders = True
    trader_instance.simulate = False
    oco_group = personal_data.OneCancelsTheOtherOrderGroup("name",
                                                           exchange_manager.exchange_personal_data.orders_manager)
    assert isinstance(oco_group.active_order_swap_strategy, personal_data.StopFirstActiveOrderSwapStrategy)
    stop_loss = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss.update(
        price=decimal.Decimal(10),
        quantity=decimal.Decimal(1),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=oco_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss, decimal.Decimal(10), False),
    )
    sell_limit = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit.update(
        price=decimal.Decimal(20),
        quantity=decimal.Decimal(1),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=oco_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit, decimal.Decimal(20), True),
    )
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    assert portfolio["BTC"].available == decimal.Decimal(10)
    for order in [stop_loss, sell_limit]:
        order.exchange_manager.is_backtesting = True  # force update_order_status
        await order.initialize()
        await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(
            order
        )
        if order is stop_loss:
            # stop sell limit quantity is not removed from available but stop is
            assert portfolio["BTC"].available == decimal.Decimal(9)
        else:
            assert portfolio["BTC"].available == decimal.Decimal(9)
    assert stop_loss.is_active is True
    assert order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False) == [sell_limit]
    with (mock.patch.object(exchange_manager.exchange, "cancel_order", mock.AsyncMock(return_value=enums.OrderStatus.CANCELED)) as cancel_order_mock):
        assert await oco_group.adapt_before_order_becoming_active(sell_limit) == ([stop_loss], oco_group._reverse_active_swap)
        cancel_order_mock.assert_called_once()
        assert stop_loss.is_active is False
        assert not stop_loss.is_cleared()
        assert not sell_limit.is_cleared()
        # stop loss is now inactive as well, could be to update sell limit to come active (done in strategy)
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert len(inactive_orders) == 2
        assert stop_loss in inactive_orders and sell_limit in inactive_orders


async def test_oco_reverse_active_swap(oco_group):
    order_1 = mock.Mock()
    order_2 = mock.Mock()
    to_activate_orders = [order_1, order_2]
    with mock.patch.object(oco_group, "adapt_before_order_becoming_active", mock.AsyncMock()) as adapt_before_order_becoming_active_mock, \
        mock.patch.object(order_util, "create_as_active_order_on_exchange", mock.AsyncMock()) as create_as_active_order_on_exchange_mock:
        await oco_group._reverse_active_swap("plop", to_activate_orders)
        assert adapt_before_order_becoming_active_mock.call_count == 2
        assert create_as_active_order_on_exchange_mock.call_count == 2
        assert adapt_before_order_becoming_active_mock.mock_calls[0].args == (order_1,)
        assert adapt_before_order_becoming_active_mock.mock_calls[1].args == (order_2,)
        assert create_as_active_order_on_exchange_mock.mock_calls[0].args == (order_1, False)
        assert create_as_active_order_on_exchange_mock.mock_calls[1].args == (order_2, False)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import octobot_trading.personal_data
import octobot_trading.enums as enums

import pytest
from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import sell_limit_order, buy_limit_order, buy_market_order, sell_market_order
import tests.personal_data.orders.states as states

pytestmark = pytest.mark.asyncio


async def test_on_order_refresh_successful(sell_limit_order):
    sell_limit_order.status = enums.OrderStatus.CANCELED
    sell_limit_order.exchange_manager.is_backtesting = True
    await sell_limit_order.initialize()
    await sell_limit_order.state.on_refresh_successful()
    assert sell_limit_order.is_cancelled()
    sell_limit_order.clear()


async def test_initialize_without_kwargs(sell_limit_order):
    await states.inner_test_initialize_without_kwargs(sell_limit_order, enums.OrderStatus.CANCELED, 'on_close')


async def test_initialize_with_kwargs(sell_limit_order):
    await states.inner_test_initialize_with_kwargs(sell_limit_order, enums.OrderStatus.CANCELED, 'on_close')


async def test_constructor_real_order_with_pending_cancel_status(buy_limit_order):
    # with PENDING_CANCEL status
    buy_limit_order.status = enums.OrderStatus.PENDING_CANCEL
    buy_limit_order.simulated = False
    state = octobot_trading.personal_data.CancelOrderState(buy_limit_order, True)
    assert state.state is enums.OrderStates.CANCELING
    assert state.is_pending()
    assert not state.is_canceled()
    if not os.getenv('CYTHON_IGNORE'):
        assert state.is_status_pending()
        assert not state.is_status_cancelled()


async def test_constructor_simulated_order_with_pending_cancel_status(buy_limit_order):
    buy_limit_order.status = enums.OrderStatus.PENDING_CANCEL
    buy_limit_order.simulated = True
    state = octobot_trading.personal_data.CancelOrderState(buy_limit_order, True)
    assert state.state is enums.OrderStates.CANCELED
    assert not state.is_pending()
    assert state.is_canceled()
    if not os.getenv('CYTHON_IGNORE'):
        assert not state.is_status_pending()
        assert state.is_status_cancelled()


async def test_constructor_with_cancelled_status(sell_limit_order):
    # with CANCELED status
    sell_limit_order.status = enums.OrderStatus.CANCELED
    state = octobot_trading.personal_data.CancelOrderState(sell_limit_order, True)
    assert state.state is enums.OrderStates.CANCELED
    assert not state.is_pending()
    assert state.is_canceled()
    if not os.getenv('CYTHON_IGNORE'):
        assert not state.is_status_pending()
        assert state.is_status_cancelled()


async def test_constructor_with_expired_status(buy_market_order):
    # with EXPIRED status
    buy_market_order.status = enums.OrderStatus.EXPIRED
    state = octobot_trading.personal_data.CancelOrderState(buy_market_order, True)
    assert state.state is enums.OrderStates.CANCELED
    assert not state.is_pending()
    assert state.is_canceled()
    if not os.getenv('CYTHON_IGNORE'):
        assert not state.is_status_pending()
        assert state.is_status_cancelled()


async def test_constructor_with_rejected_status(sell_market_order):
    # with REJECTED status
    sell_market_order.status = enums.OrderStatus.REJECTED
    state = octobot_trading.personal_data.CancelOrderState(sell_market_order, True)
    assert state.state is enums.OrderStates.CANCELED
    assert not state.is_pending()
    assert state.is_canceled()
    if not os.getenv('CYTHON_IGNORE'):
        assert not state.is_status_pending()
        assert state.is_status_cancelled()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import contextlib
try:
    from aiohttp_socks import ProxyConnectionError
except ImportError:
    # local mock in case aiohttp_socks is not available
    class ProxyConnectionError(Exception):
        pass
import os
import ssl
import aiohttp
import copy
import logging
import typing
import ccxt
import ccxt.pro as ccxt_pro
import ccxt.async_support as async_ccxt

import octobot_commons
import octobot_commons.time_frame_manager as time_frame_manager
import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.logging as commons_logging
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums
import octobot_trading.exchanges.connectors.ccxt.ccxt_clients_cache as ccxt_clients_cache
import octobot_trading.exchanges.config.proxy_config as proxy_config_import
import octobot_trading.exchanges.config.exchange_credentials_data as exchange_credentials_data
import octobot_trading.exchanges.util.exchange_util as exchange_util


def create_client(
    exchange_class, exchange_manager, logger, options, headers,
    additional_config, should_authenticate, unauthenticated_exchange_fallback=None,
    keys_adapter=None, allow_request_counter: bool = True
):
    """
    Exchange instance creation
    :return: the created ccxt (pro, async or sync) client
    """
    is_authenticated = False
    should_be_authenticated_exchange = should_authenticate and not exchange_manager.is_backtesting
    if not exchange_manager.exchange_only:
        # avoid logging version on temporary exchange_only exchanges
        exchange_type = exchange_util.get_exchange_type(exchange_manager)
        logger.info(
            f"Creating {'' if should_be_authenticated_exchange else 'un'}authenticated {exchange_class.__name__} "
            f"{exchange_type.name} exchange with ccxt in version {ccxt.__version__}"
        )
    if exchange_manager.ignore_config or exchange_manager.check_config(exchange_manager.exchange_name):
        try:
            creds: exchange_credentials_data.ExchangeCredentialsData = (
                exchange_manager.get_exchange_credentials(exchange_manager.exchange_name)
            )
            if keys_adapter:
                creds = keys_adapter(creds)
            if not (creds.has_credentials()) and not exchange_manager.is_simulated and not exchange_manager.ignore_config:
                logger.warning(f"No exchange API key set for {exchange_manager.exchange_name}. "
                               f"Enter your account details to enable real trading on this exchange.")
            if should_be_authenticated_exchange:
                client = instantiate_exchange(
                    exchange_class,
                    _get_client_config(
                        exchange_class, options, headers, additional_config, creds,
                    ),
                    exchange_manager.exchange_name,
                    exchange_manager.proxy_config,
                    allow_request_counter=allow_request_counter
                )
                is_authenticated = True
                if exchange_manager.check_credentials:
                    client.check_required_credentials()
            else:
                client = instantiate_exchange(
                    exchange_class,
                    _get_client_config(exchange_class, options, headers, additional_config),
                    exchange_manager.exchange_name,
                    exchange_manager.proxy_config,
                    allow_request_counter=allow_request_counter
                )
        except (ccxt.AuthenticationError, Exception) as e:
            if unauthenticated_exchange_fallback is None:
                return get_unauthenticated_exchange(
                    exchange_class, options, headers, additional_config,
                    exchange_manager.exchange_name, exchange_manager.proxy_config
                ), False
            return unauthenticated_exchange_fallback(e), False
    else:
        client = get_unauthenticated_exchange(
            exchange_class, options, headers, additional_config,
            exchange_manager.exchange_name, exchange_manager.proxy_config
        )
        logger.error("configuration issue: missing login information !")
    client.logger.setLevel(logging.INFO)
    return client, is_authenticated


async def close_client(client):
    await client.close()
    client.markets = {}
    client.markets_by_id = {}
    client.ids = []
    client.last_json_response = {}
    client.last_http_response = ""
    client.last_response_headers = {}
    client.markets_loading = None
    client.currencies = {}
    client.baseCurrencies = {}
    client.quoteCurrencies = {}
    client.currencies_by_id = {}
    client.codes = []
    client.symbols = {}
    client.accounts = []
    client.accounts_by_id = {}
    client.ohlcvs = {}
    client.trades = {}
    client.orderbooks = {}


def get_unauthenticated_exchange(
    exchange_class, options, headers, additional_config, identifier: str, proxy_config: proxy_config_import.ProxyConfig
) -> async_ccxt.Exchange:
    return instantiate_exchange(
        exchange_class,
        _get_client_config(exchange_class, options, headers, additional_config),
        identifier,
        proxy_config
    )


def instantiate_exchange(
    exchange_class, config: dict, identifier: str, proxy_config: proxy_config_import.ProxyConfig,
    allow_request_counter: bool = True
) -> async_ccxt.Exchange:
    client = exchange_class(config)
    _use_proxy_if_necessary(client, proxy_config)
    if constants.ENABLE_CCXT_REQUESTS_COUNTER and allow_request_counter:
        if proxy_config.socks_proxy or proxy_config.socks_proxy_callback:
            commons_logging.get_logger(__name__).error("socks proxy and request counter can't yet be used together.")
        else:
            _use_request_counter(identifier, client, proxy_config)
    return client


def set_sandbox_mode(exchange_connector, is_sandboxed):
    try:
        if exchange_connector.exchange_manager.exchange.uses_demo_trading_instead_of_sandbox():
            exchange_connector.client.enable_demo_trading(is_sandboxed)
        else:
            exchange_connector.client.set_sandbox_mode(is_sandboxed)
    except ccxt.NotSupported as e:
        default_type = exchange_connector.client.options.get('defaultType', None)
        additional_info = f" in type {default_type}" if default_type else ""
        exchange_connector.logger.warning(f"{exchange_connector.name} does not support sandboxing {additional_info}: {e}")
        # raise exception to stop this exchange and prevent dealing with a real funds exchange
        raise e
    return None


@contextlib.contextmanager
def filtered_fetched_markets(client, market_filter: typing.Callable[[dict], bool]):
    origin_fetch_markets = client.fetch_markets

    async def _filted_fetched_markets(*args, **kwargs):
        all_markets = await origin_fetch_markets(*args, **kwargs)
        filtered_markets = [
            market 
            for market in all_markets
            if market_filter(market)
        ]
        commons_logging.get_logger(__name__).info(
            f"Keeping {len(filtered_markets)} out of {len(all_markets)} fetched markets"
        )
        return filtered_markets
    try:
        client.fetch_markets = _filted_fetched_markets
        yield
    finally:
        client.fetch_markets = origin_fetch_markets


def load_markets_from_cache(client, authenticated_cache: bool, market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None):
    client.set_markets(
        market
        for market in ccxt_clients_cache.get_exchange_parsed_markets(ccxt_clients_cache.get_client_key(client, authenticated_cache))
        if market_filter is None or market_filter(market)
    )


def set_markets_cache(client, authenticated_cache: bool):
    if client.markets:
        ccxt_clients_cache.set_exchange_parsed_markets(
            ccxt_clients_cache.get_client_key(client, authenticated_cache), copy.deepcopy(list(client.markets.values()))
        )


def get_ccxt_client_login_options(exchange_manager):
    """
    :return: ccxt client login option dict, can be overwritten to custom exchange login
    """
    if exchange_manager.is_future:
        return {'defaultType': 'future'}
    if exchange_manager.is_margin:
        return {'defaultType': 'margin'}
    return {'defaultType': 'spot'}


def get_symbols(client, active_only) -> set[str]:
    try:
        if active_only:
            return set(
                symbol
                for symbol in client.symbols
                if client.markets.get(symbol, {}).get(
                    enums.ExchangeConstantsMarketStatusColumns.ACTIVE.value, True
                ) in (True, None)
            )
        return set(client.symbols)
    except (AttributeError, TypeError):
        # ccxt exchange load_markets failed
        return set()


def get_time_frames(client):
    try:
        if isinstance(client, ccxt_pro.Exchange):
            # ccxt pro exchanges might have different timeframes in options
            options_time_frames = client.safe_value(client.options, 'timeframes')
            if options_time_frames:
                values = set([
                    time_frame
                    for time_frame in options_time_frames
                    if time_frame_manager.is_time_frame(time_frame)
                ])
                if values:
                    return values
        # use normal client timeframes (values of rest exchange)
        return set(client.timeframes)
    except (AttributeError, TypeError):
        # ccxt exchange describe() is invalid
        return set()


def get_exchange_pair(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["id"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_pair_cryptocurrency(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["base"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_contract_size(client, pair) -> float:
    return get_market_status_contract_size(client.markets[pair])


def get_market_status_contract_size(market_status: dict) :
    return market_status[ccxt_enums.ExchangeConstantsMarketStatusCCXTColumns.CONTRACT_SIZE.value]


def get_fees(market_status) -> dict:
    return {
        enums.ExchangeConstantsMarketPropertyColumns.TAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.TAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.MAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.MAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.FEE.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.FEE.value,
                              constants.CONFIG_DEFAULT_FEES)
    }


def add_headers(client, headers_dict):
    """
    Add new headers to ccxt client
    :param headers_dict: the additional header keys and values as dict
    """
    for header_key, header_value in headers_dict.items():
        client.headers[header_key] = header_value


def add_options(client, options_dict):
    """
    Add new options to ccxt client
    :param options_dict: the additional option keys and values as dict
    """
    for option_key, option_value in options_dict.items():
        client.options[option_key] = option_value


def converted_ccxt_common_errors(f):
    async def converted_ccxt_common_errors_wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except ccxt.RateLimitExceeded as err:
            raise errors.RateLimitExceeded(err) from err
        except ccxt.NotSupported as err:
            raise errors.NotSupported(err) from err
    return converted_ccxt_common_errors_wrapper


def _use_proxy_if_necessary(client, proxy_config: proxy_config_import.ProxyConfig):
    client.aiohttp_trust_env = proxy_config.aiohttp_trust_env
    if proxy_config.http_proxy:
        client.http_proxy = proxy_config.http_proxy
    if proxy_config.http_proxy_callback:
        client.http_proxy_callback = proxy_config.http_proxy_callback
    if proxy_config.https_proxy:
        client.https_proxy = proxy_config.https_proxy
    if proxy_config.https_proxy_callback:
        client.https_proxy_callback = proxy_config.https_proxy_callback
    if proxy_config.socks_proxy:
        client.socks_proxy = proxy_config.socks_proxy
    if proxy_config.socks_proxy_callback:
        client.socks_proxy_callback = proxy_config.socks_proxy_callback
    if proxy_config.socks_proxy or proxy_config.socks_proxy_callback:
        # rewrite of async_ccxt.exchange.client.fetch() ProxyConnector creation
        _init_ccxt_client_session_requirements(client)
        proxy_url = proxy_config.get_proxy_url()
        if (client.socks_proxy_sessions is None):
            client.socks_proxy_sessions = {}
        if (proxy_url not in client.socks_proxy_sessions):
            try:
                import aiohttp_socks
                previous_aiohttp_socks_connector = client.aiohttp_socks_connector
                client.aiohttp_socks_connector = aiohttp_socks.ProxyConnector.from_url(
                    proxy_url,
                    # extra args copied from self.open()
                    ssl=client.ssl_context,
                    loop=client.asyncio_loop,
                    enable_cleanup_closed=True
                )
                client.socks_proxy_sessions[proxy_url] = aiohttp.ClientSession(
                    loop=client.asyncio_loop, connector=client.aiohttp_socks_connector,
                    trust_env=client.aiohttp_trust_env
                )
                asyncio.create_task(_close_previous_session_and_connector(
                    None, previous_aiohttp_socks_connector
                ))
            except ImportError as err:
                raise ImportError(
                    "The aiohttp_socks python library is not installed and is required to use a socks proxy"
                ) from err


def _init_ccxt_client_session_requirements(client):
    # from async_ccxt.exchange.client.open()
    if client.asyncio_loop is None:
        client.asyncio_loop = asyncio.get_running_loop()
        client.throttler.loop = client.asyncio_loop

    if client.ssl_context is None:
        # Create our SSL context object with our CA cert file
        client.ssl_context = ssl.create_default_context(cafile=client.cafile) if client.verify else client.verify



def _get_client_config(
    exchange_class, options, headers, additional_config, creds: exchange_credentials_data.ExchangeCredentialsData=None
):
    if creds and creds.auth_token:
        headers["Authorization"] = f"{creds.auth_token_header_prefix or ''}{creds.auth_token}"
    config = {
        'verbose': constants.ENABLE_CCXT_VERBOSE,
        'enableRateLimit': constants.ENABLE_CCXT_RATE_LIMIT,
        'timeout': constants.DEFAULT_REQUEST_TIMEOUT,
        'options': options,
        'headers': headers,
        'timeout_on_exit': constants.CCXT_TIMEOUT_ON_EXIT_MS,
    }
    if creds:
        if creds.api_key is not None:
            config['apiKey'] = creds.api_key
        if creds.secret is not None:
            config['secret'] = creds.secret
        if creds.password is not None:
            config['password'] = creds.password
        if creds.uid is not None:
            config['uid'] = creds.uid
        if creds.wallet_address is not None:
            config['walletAddress'] = creds.wallet_address
        if creds.private_key is not None:
            config['privateKey'] = creds.private_key
    config.update({**get_custom_domain_config(exchange_class), **(additional_config or {})})
    return config


def get_custom_domain_config(exchange_class):
    old, new = _get_replaced_custom_domains(exchange_class)
    if not (old and new):
        return {}
    if url_config := exchange_class().describe()[ccxt_enums.ExchangeColumns.URLS.value]:
        commons_logging.get_logger(__name__).info(
            f"Using custom domain for {exchange_class.__name__}: {old} is replaced by {new}, hostname has been updated"
        )
        return {
            ccxt_enums.ExchangeColumns.URLS.value: _get_patched_url_config(url_config, old, new),
            ccxt_enums.ExchangeColumns.HOSTNAME.value: new
        }
    return {}


def _get_replaced_custom_domains(exchange_class):
    identifier = exchange_class.__name__.upper()
    if custom_domain := os.getenv(f"{identifier}_CUSTOM_DOMAIN"):
        split = custom_domain.split(":")
        if len(split) == 2:
            return split[0], split[1]
        else:
            commons_logging.get_logger(__name__).error(
                f"Invalid {identifier} custom domain config. Expected syntax is to_replace_domain:updated_domain "
                f"Example: MEXC_CUSTOM_DOMAIN=mexc.com:mexc.co"
            )
    return None, None


def _get_patched_url_config(url_config: dict, old: str, new: str):
    updated_config = {}
    for key, val in url_config.items():
        if isinstance(val, dict):
            updated_config[key] = _get_patched_url_config(val, old, new)
        elif isinstance(val, str):
            updated_config[key] = val.replace(old, new)
        else:
            updated_config[key] = val
    return updated_config


async def _close_previous_session_and_connector(session, connector):
    if connector is not None:
        await connector.close()
    if session is not None:
        await session.close()


def _use_request_counter(
    identifier: str, ccxt_client: async_ccxt.Exchange, proxy_config: proxy_config_import.ProxyConfig
):
    """
    Replaces the given exchange async session by an aiohttp_util.CounterClientSession
    WARNING: should only be called right after creating the exchange and on the same async loop as
    the one the exchange will be using (to avoid interrupting open requests from current session and loop errors)
    """
    # use session with request counter
    try:
        # 1. create ssl context and other required elements if necessary
        ccxt_client.open()
        previous_session = ccxt_client.session
        previous_connector = ccxt_client.tcp_connector
        # 2. create patched session using the same params as a normal one
        # same as in ccxt.async_support.exchange.py#open()
        # connector = aiohttp.TCPConnector(ssl=self.ssl_context, loop=self.asyncio_loop, enable_cleanup_closed=True)
        ccxt_client.tcp_connector = aiohttp.TCPConnector(
            ssl=ccxt_client.ssl_context, loop=ccxt_client.asyncio_loop,
            enable_cleanup_closed=True
        )
        counter_session = aiohttp_util.CounterClientSession(
            identifier,
            loop=ccxt_client.asyncio_loop,
            connector=ccxt_client.tcp_connector,
            trust_env=previous_session.trust_env,
        )
        # 3. replace session
        ccxt_client.session = counter_session
        # 4. close replaced session in task to avoid making this function a coroutine
        asyncio.create_task(_close_previous_session_and_connector(previous_session, previous_connector))
        commons_logging.get_logger(__name__).info(f"Request counter enabled for {identifier}")
    except Exception as err:
        commons_logging.get_logger(__name__).exception(
            err, True, f"Error when initializing {identifier} request counter: {err}"
        )


def ccxt_exchange_class_factory(exchange_name):
    return getattr(async_ccxt, exchange_name)


def reraise_with_proxy_prefix_if_relevant(
    ccxt_connector, cause_error: Exception, raised_error: typing.Optional[Exception]
):
    was_proxied, last_proxied_request_url = was_latest_request_proxied(ccxt_connector)
    if was_proxied:
        raised = raised_error or cause_error
        raise raised.__class__(f"[Proxied] {raised} [URL: {last_proxied_request_url}]") from cause_error


def was_latest_request_proxied(ccxt_connector) -> (bool, str):
    if not (
        ccxt_connector.exchange_manager.proxy_config
        and ccxt_connector.exchange_manager.proxy_config.get_last_proxied_request_url
    ):
        return False, ""
    last_proxied_request_url = ccxt_connector.exchange_manager.proxy_config.get_last_proxied_request_url()
    last_client_request_url = ccxt_connector.client.last_request_url
    # if last requests are matching: it was proxied
    if last_proxied_request_url:
        url_without_param = last_proxied_request_url.split("?")[0]
        return last_proxied_request_url == last_client_request_url, url_without_param
    return False, ""


def get_proxy_error_class(proxy_error: Exception):
    if _is_retriable_proxy_error(proxy_error):
        return errors.RetriableExchangeProxyError
    return errors.ExchangeProxyError


def _is_retriable_proxy_error(proxy_error: Exception) -> bool:
    str_err = str(proxy_error)
    for desc in constants.RETRIABLE_EXCHANGE_PROXY_ERRORS_DESC:
        if desc in str_err:
            return True
    return False


def get_proxy_error_if_any(ccxt_connector, error: Exception) -> typing.Optional[Exception]:
    if not ccxt_connector.exchange_manager.proxy_config:
        return None
    max_depth = 10
    depth = 1
    cause_error = error
    while cause_error and depth < max_depth:
        if isinstance(cause_error, (
            # Proxy errors
            aiohttp.ClientProxyConnectionError, aiohttp.ClientHttpProxyError, ProxyConnectionError
        )) or (
            # Connector error with the configured proxy host in description
            isinstance(cause_error, aiohttp.ClientConnectorError)
            and ccxt_connector.exchange_manager.proxy_config.proxy_host in str(cause_error)
        ):
            return cause_error
        depth += 1
        cause_error = getattr(cause_error, "__cause__", None)
    return None


def fix_client_missing_markets_fees(
    client: async_ccxt.Exchange, reloaded_markets: bool, confirmed_fees: dict
):
    fees_by_pair_suffix = {}
    update_confirmed_fees = reloaded_markets or not confirmed_fees
    for symbol, market in client.markets.items():
        pair_suffix = _get_market_symbol_suffix(symbol)
        maker_fees = market.get(enums.ExchangeConstantsMarketPropertyColumns.MAKER.value)
        taker_fees = market.get(enums.ExchangeConstantsMarketPropertyColumns.TAKER.value)
        if taker_fees is not None and maker_fees is not None:
            if update_confirmed_fees:
                # only update cached fees if reload is True or if it's the first call 
                # (avoid storing inferred fees)
                confirmed_fees[symbol] = (maker_fees, taker_fees)
            if (
                # be pessimistic and use the highest fees of each market
                pair_suffix not in fees_by_pair_suffix or fees_by_pair_suffix[pair_suffix][0] < maker_fees or fees_by_pair_suffix[pair_suffix][1] < taker_fees
            ):
                fees_by_pair_suffix[pair_suffix] = (maker_fees, taker_fees)
    for symbol, market in client.markets.items():
        # use other similar market fees if missing
        try:
            pair_suffix = _get_market_symbol_suffix(symbol)
            if market[enums.ExchangeConstantsMarketPropertyColumns.MAKER.value] is None or market[enums.ExchangeConstantsMarketPropertyColumns.TAKER.value] is None:
                has_cached_fees = symbol in confirmed_fees
                market[enums.ExchangeConstantsMarketPropertyColumns.MAKER.value] = (
                    market[enums.ExchangeConstantsMarketPropertyColumns.MAKER.value] or (
                        (confirmed_fees[symbol] if has_cached_fees else fees_by_pair_suffix[pair_suffix])[0]
                    )
                )
                market[enums.ExchangeConstantsMarketPropertyColumns.TAKER.value] = (
                    market[enums.ExchangeConstantsMarketPropertyColumns.TAKER.value] or (
                        (confirmed_fees[symbol] if has_cached_fees else fees_by_pair_suffix[pair_suffix])[1]
                    )
                )
                commons_logging.get_logger("fix_missing_markets_fees").info(
                    f"Fixed missing {symbol} fees using {'cached fees' if has_cached_fees else str(pair_suffix)} fees: {fees_by_pair_suffix[pair_suffix]}" 
                )
        except KeyError as err:
            commons_logging.get_logger("fix_missing_markets_fees").error(
                f"Failed to fix missing market fees for {symbol}: {err}"
            )
        except Exception as err:
            commons_logging.get_logger("fix_missing_markets_fees").exception(
                err, True, f"Unexpected error when fixing missing market fees for {symbol}: {err}"
            )


def _get_market_symbol_suffix(symbol):
    return symbol[symbol.index(octobot_commons.MARKET_SEPARATOR):]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os

import pytest

import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.exchange_data.contracts as contracts

from tests import event_loop

pytestmark = pytest.mark.asyncio

DEFAULT_SYMBOL = "BTC/USDT"


async def test_is_isolated():
    contract = contracts.MarginContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        current_leverage=decimal.Decimal(10))
    assert contract.is_isolated()
    contract = contracts.MarginContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.CROSS,
        current_leverage=decimal.Decimal(10))
    assert not contract.is_isolated()


async def test_set_current_leverage():
    contract = contracts.MarginContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        maximum_leverage=None,
        current_leverage=decimal.Decimal(10))
    assert contract.current_leverage == decimal.Decimal(10)
    assert contract.maximum_leverage is None
    contract.set_current_leverage(decimal.Decimal(50))
    assert contract.current_leverage == decimal.Decimal(50)
    contract = contracts.MarginContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        maximum_leverage=decimal.Decimal(100),
        current_leverage=decimal.Decimal(10))
    assert contract.current_leverage == decimal.Decimal(10)
    contract.set_current_leverage(decimal.Decimal(50))
    assert contract.current_leverage == decimal.Decimal(50)


async def test_set_margin_type():
    contract = contracts.MarginContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        current_leverage=decimal.Decimal(10))
    assert contract.margin_type == enums.MarginType.ISOLATED
    contract.set_margin_type(is_isolated=True)
    assert contract.margin_type == enums.MarginType.ISOLATED
    contract.set_margin_type(is_cross=False)
    assert contract.margin_type == enums.MarginType.ISOLATED
    contract.set_margin_type(is_isolated=False)
    assert contract.margin_type == enums.MarginType.CROSS
    contract.set_margin_type(is_cross=False)
    assert contract.margin_type == enums.MarginType.ISOLATED
    contract.set_margin_type(is_cross=True)
    assert contract.margin_type == enums.MarginType.CROSS




#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import dataclasses
import decimal
import typing

import octobot_commons.logging as commons_logging
import octobot_commons.constants as commons_constants
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants


@dataclasses.dataclass
class SubPortfolioData:
    bot_id: typing.Optional[str]
    portfolio_id: typing.Optional[str]
    priority_key: float
    content: dict[str, dict[str, decimal.Decimal]]  # current content of the sub-portfolio
    unit: typing.Optional[str]
    allowed_filling_assets: list[str] = dataclasses.field(default_factory=list) # assets to use to fill missing funds
    forbidden_filling_assets: list[str] = dataclasses.field(default_factory=list) # assets NOT to use to fill missing funds
    # deltas to be applied on top of the current content of the sub-portfolio from get_content_after_deltas()
    funds_deltas: dict[str, dict[str, decimal.Decimal]] = dataclasses.field(default_factory=dict)
    # funds that are missing from this portfolio. Populated after portfolio has been resolved
    missing_funds: dict[str, decimal.Decimal] = dataclasses.field(default_factory=dict)
    # funds that are locked (maybe in orders) from this portfolio
    locked_funds_by_asset: dict[str, decimal.Decimal] = dataclasses.field(default_factory=dict)


    def get_content_after_deltas(self) -> dict[str, dict[str, decimal.Decimal]]:
        """
        Computes the updated portfolio from delta available and total
        """
        updated_content = copy.deepcopy(self.content)
        for asset, delta_values in self.funds_deltas.items():
            if asset in updated_content:
                updated_content[asset][commons_constants.PORTFOLIO_TOTAL] += (
                    delta_values[commons_constants.PORTFOLIO_TOTAL]
                )
                updated_content[asset][commons_constants.PORTFOLIO_AVAILABLE] += (
                    delta_values[commons_constants.PORTFOLIO_AVAILABLE]
                )
            else:
                updated_content[asset] = {
                    commons_constants.PORTFOLIO_TOTAL: delta_values[commons_constants.PORTFOLIO_TOTAL],
                    commons_constants.PORTFOLIO_AVAILABLE: delta_values[commons_constants.PORTFOLIO_AVAILABLE],
                }
            for key in (commons_constants.PORTFOLIO_TOTAL, commons_constants.PORTFOLIO_AVAILABLE):
                if updated_content[asset][key] < constants.ZERO:
                    commons_logging.get_logger(__name__).warning(
                        f"{asset} portfolio {key} value: {updated_content[asset][key]} is < 0. Replacing with zero."
                    )
                    updated_content[asset][key] = constants.ZERO
        return updated_content

    def get_content_from_total_deltas_and_locked_funds(self) -> dict[str, dict[str, decimal.Decimal]]:
        """
        Only uses PORTFOLIO_TOTAL deltas and computes available values from locked_funds_by_asset
        """
        return get_content_with_available_after_deltas(
            self.content, self.funds_deltas, self.locked_funds_by_asset
        )

    def is_similar_to(self, other) -> bool:
        return (
            other
            and personal_data.filter_empty_values(self.content) == personal_data.filter_empty_values(other.content)
        )


def get_content_with_available_after_deltas(
    content: dict[str, dict[str, decimal.Decimal]], 
    deltas: dict[str, dict[str, decimal.Decimal]],
    locked_funds_by_asset: typing.Optional[dict[str, decimal.Decimal]] = None,
) -> dict[str, dict[str, decimal.Decimal]]:
    updated_content = get_content_after_deltas(content, deltas, apply_available_deltas=False)
    update_available_considering_locked_funds(updated_content, locked_funds_by_asset)
    return updated_content


def get_content_after_deltas(
    content: dict[str, dict[str, decimal.Decimal]], deltas: dict[str, dict[str, decimal.Decimal]], apply_available_deltas: bool = False
) -> dict[str, dict[str, decimal.Decimal]]:
    updated_content = copy.deepcopy(content)
    for asset, delta_values in deltas.items():
        if asset in updated_content:
            updated_content[asset][commons_constants.PORTFOLIO_TOTAL] += (
                delta_values[commons_constants.PORTFOLIO_TOTAL]
            )
            if apply_available_deltas:
                updated_content[asset][commons_constants.PORTFOLIO_AVAILABLE] += (
                    delta_values[commons_constants.PORTFOLIO_AVAILABLE]
                )
        else:
            updated_content[asset] = {
                commons_constants.PORTFOLIO_TOTAL: delta_values[commons_constants.PORTFOLIO_TOTAL]
            }
            if apply_available_deltas:
                updated_content[asset][commons_constants.PORTFOLIO_AVAILABLE] = (
                    delta_values[commons_constants.PORTFOLIO_AVAILABLE]
                )
    return updated_content


def update_available_considering_locked_funds(
    content: typing.Optional[dict[str, dict[str, decimal.Decimal]]],
    locked_funds_by_asset: typing.Optional[dict[str, decimal.Decimal]] = None,
) -> None:
    locked_funds_by_asset = locked_funds_by_asset or {}
    for asset, values in content.items():
        locked_funds = locked_funds_by_asset.get(asset, constants.ZERO)
        if locked_funds == constants.ZERO:
            values[commons_constants.PORTFOLIO_AVAILABLE] = values[commons_constants.PORTFOLIO_TOTAL]
        else:
            if locked_funds > values[commons_constants.PORTFOLIO_TOTAL]:
                delta = locked_funds - values[commons_constants.PORTFOLIO_TOTAL]
                error_details = (
                    f"{asset} available funds after applying {locked_funds} "
                    f"locked funds to total holdings of {values[commons_constants.PORTFOLIO_TOTAL]} (delta={delta * decimal.Decimal('-1')}). "
                    f"Available value will be set to 0."
                )
                if delta < values[commons_constants.PORTFOLIO_TOTAL] * decimal.Decimal("0.05"):
                    # delta < 5% of total: log warning: this is likely due to locked funds for open order fees, just warn
                    commons_logging.get_logger(__name__).warning(f"Tiny negative {error_details}")
                else:
                    # large negative: this should not happen,log error
                    commons_logging.get_logger(__name__).error(f"Unexpected: large negative {error_details}")
            values[commons_constants.PORTFOLIO_AVAILABLE] = values[commons_constants.PORTFOLIO_TOTAL] - locked_funds



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import dataclasses

import octobot_commons.signals as signals
import octobot_trading.enums as trading_enums
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as trading_constants


def create_order_signal_content(
    order, action, strategy, exchange_manager,
    target_amount=None,
    target_position=None,
    updated_target_amount=None,
    updated_target_position=None,
    updated_limit_price=trading_constants.ZERO,
    updated_stop_price=trading_constants.ZERO,
    updated_current_price=trading_constants.ZERO,
) -> dict:
    # only use order.order_id to identify orders in signals (exchange_order_id is local and never shared)
    return {
        trading_enums.TradingSignalCommonsAttrs.ACTION.value: action.value,
        trading_enums.TradingSignalOrdersAttrs.STRATEGY.value: strategy,
        trading_enums.TradingSignalOrdersAttrs.SYMBOL.value: order.symbol,
        trading_enums.TradingSignalOrdersAttrs.EXCHANGE.value: exchange_manager.exchange_name,
        trading_enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: exchanges.get_exchange_type(exchange_manager).value,
        trading_enums.TradingSignalOrdersAttrs.SIDE.value: order.side.value,
        trading_enums.TradingSignalOrdersAttrs.TRIGGER_ABOVE.value: order.trigger_above,
        trading_enums.TradingSignalOrdersAttrs.TYPE.value: order.order_type.value,
        trading_enums.TradingSignalOrdersAttrs.QUANTITY.value: float(order.origin_quantity),
        trading_enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: target_amount
        if updated_target_amount is None else updated_target_amount,
        trading_enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: target_position
        if updated_target_position is None else updated_target_position,
        trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: updated_target_amount,
        trading_enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: updated_target_position,
        trading_enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: float(updated_limit_price)
        if updated_limit_price else float(order.origin_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: float(updated_limit_price),
        trading_enums.TradingSignalOrdersAttrs.STOP_PRICE.value: float(updated_stop_price)
        if updated_stop_price else float(order.origin_stop_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: float(updated_stop_price),
        trading_enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: float(updated_current_price)
        if updated_current_price else float(order.created_last_price),
        trading_enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: float(updated_current_price),
        trading_enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: order.reduce_only,
        trading_enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        trading_enums.TradingSignalOrdersAttrs.GROUP_ID.value:
            None if order.order_group is None else order.order_group.name,
        trading_enums.TradingSignalOrdersAttrs.GROUP_TYPE.value:
            None if order.order_group is None else order.order_group.__class__.__name__,
        trading_enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TYPE.value: \
            order.order_group.active_order_swap_strategy.__class__.__name__ if order.order_group else None,
        trading_enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TIMEOUT.value: \
            order.order_group.active_order_swap_strategy.swap_timeout if order.order_group else None,
        trading_enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TRIGGER_CONFIG.value: \
            order.order_group.active_order_swap_strategy.trigger_price_configuration if order.order_group else None,
        trading_enums.TradingSignalOrdersAttrs.TAG.value: order.tag,
        trading_enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: order.associated_entry_ids,
        trading_enums.TradingSignalOrdersAttrs.UPDATE_WITH_TRIGGERING_ORDER_FEES.value:
            order.update_with_triggering_order_fees,
        trading_enums.TradingSignalOrdersAttrs.ORDER_ID.value: order.order_id,
        trading_enums.TradingSignalOrdersAttrs.TRAILING_PROFILE_TYPE.value:
            order.trailing_profile.get_type().value if order.trailing_profile else None,
        trading_enums.TradingSignalOrdersAttrs.TRAILING_PROFILE.value:
            order.trailing_profile.to_dict() if order.trailing_profile else None,
        trading_enums.TradingSignalOrdersAttrs.IS_ACTIVE.value: order.is_active,
        trading_enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_PRICE.value:
            None if order.active_trigger is None else float(order.active_trigger.trigger_price),
        trading_enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_ABOVE.value: None
            if order.active_trigger is None else order.active_trigger.trigger_above,
        trading_enums.TradingSignalOrdersAttrs.CANCEL_POLICY_TYPE.value:
            order.cancel_policy.__class__.__name__ if order.cancel_policy else None,
        trading_enums.TradingSignalOrdersAttrs.CANCEL_POLICY_KWARGS.value:
            dataclasses.asdict(order.cancel_policy) if order.cancel_policy else None,
        trading_enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value:
            None if order.triggered_by is None else order.triggered_by.order_id
        if order.has_been_bundled else None,
        trading_enums.TradingSignalOrdersAttrs.CHAINED_TO.value:
            None if order.triggered_by is None else order.triggered_by.order_id,
        trading_enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
    }


def create_position_signal_content(
    action, strategy, exchange_manager, symbol, side, leverage
) -> dict:
    # only use order.order_id to identify orders in signals (exchange_order_id is local and never shared)
    return {
        trading_enums.TradingSignalCommonsAttrs.ACTION.value: action.value,
        trading_enums.TradingSignalPositionsAttrs.STRATEGY.value: strategy,
        trading_enums.TradingSignalPositionsAttrs.EXCHANGE.value: exchange_manager.exchange_name,
        trading_enums.TradingSignalPositionsAttrs.EXCHANGE_TYPE.value: exchanges.get_exchange_type(exchange_manager).value,
        trading_enums.TradingSignalPositionsAttrs.SYMBOL.value: symbol,
        trading_enums.TradingSignalPositionsAttrs.SIDE.value: side.value if side else side,
        trading_enums.TradingSignalPositionsAttrs.LEVERAGE.value: float(leverage),
    }


def get_orders_dependencies(orders: list) -> signals.SignalDependencies:
    return signals.SignalDependencies(
        [
            {
                trading_enums.TradingSignalDependencies.ORDER_ID.value: order.order_id
            }
            for order in orders
        ]
    )


def get_order_dependency(order) -> signals.SignalDependencies:
    return get_orders_dependencies([order])


def get_position_dependency(position) -> signals.SignalDependencies:
    return signals.SignalDependencies(
        [
            {
                trading_enums.TradingSignalDependencies.POSITION_SYMBOL.value: position.symbol
            }
        ]
    )


def are_dependencies_filled(signal: signals.Signal, succeeded_signals: list[signals.Signal]) -> bool:
    if not signal.dependencies:
        return True
    return signal.dependencies.is_filled_by(
        _get_signals_filled_dependencies(succeeded_signals)
    )


def _get_signals_filled_dependencies(succeeded_signals: list[signals.Signal]) -> signals.SignalDependencies:
    return signals.SignalDependencies(
        [
            filled_dependency
            for succeeded_signal in succeeded_signals
            if (filled_dependency := _get_signal_filled_dependency(succeeded_signal))
        ]
    )


def _get_signal_filled_dependency(signal: signals.Signal) -> typing.Optional[dict]:
    if signal.topic == trading_enums.TradingSignalTopics.ORDERS.value:
        if order_id := signal.content.get(
            trading_enums.TradingSignalOrdersAttrs.ORDER_ID.value, None
        ):
            return {
                trading_enums.TradingSignalDependencies.ORDER_ID.value: order_id
            }
        return None
    if signal.topic == trading_enums.TradingSignalTopics.POSITIONS.value:
        if symbol := signal.content.get(
            trading_enums.TradingSignalPositionsAttrs.SYMBOL.value, None
        ):
            return {
                trading_enums.TradingSignalDependencies.POSITION_SYMBOL.value: symbol
            }
        return None
    raise NotImplementedError(f"Unsupported signal topic: {signal.topic}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.constants as trading_constants
import octobot_trading.enums as trading_enums
import octobot_trading.errors as trading_errors


def get_position(context, symbol=None, side=trading_enums.PositionSide.BOTH.value):
    return context.exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
        symbol or context.symbol,
        _get_position_side(context, side)
    )


async def average_open_pos_entry(context, side=trading_enums.PositionSide.LONG.value):
    if context.exchange_manager.is_future:
        return get_position(context, context.symbol, side).entry_price
    elif context.exchange_manager.is_margin:
        return trading_constants.ZERO
    else:
        return trading_constants.ZERO


def _get_position_side(ctx, side):
    if is_in_one_way_position_mode(ctx):
        return trading_enums.PositionSide.BOTH

    # hedge mode
    if side == trading_enums.PositionSide.LONG.value:
        return trading_enums.PositionSide.LONG
    elif side == trading_enums.PositionSide.SHORT.value:
        return trading_enums.PositionSide.SHORT
    elif side == trading_enums.PositionSide.BOTH.value:
        raise trading_errors.InvalidArgumentError(
            "average_open_pos_entry: both sides are not implemented yet for hedged mode"
        )
    else:
        raise trading_errors.InvalidArgumentError(
            'average_open_pos_entry: side needs to be "long", "short" or "both"'
        )


def is_in_one_way_position_mode(ctx):
    return ctx.exchange_manager.exchange.get_pair_future_contract(ctx.symbol).is_one_way_position_mode()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import pytest

from tests.exchange_data import recent_trades_manager, price_events_manager

# All test coroutines will be treated as marked.
from tests.test_utils.random_numbers import random_recent_trade, random_recent_trades
from tests import event_loop

pytestmark = pytest.mark.asyncio


async def test_initialize(recent_trades_manager):
    assert not recent_trades_manager.recent_trades
    assert not recent_trades_manager.liquidations
    with pytest.raises(IndexError):
        recent_trades_manager.recent_trades[0]
    with pytest.raises(IndexError):
        recent_trades_manager.liquidations[0]


async def test_set_all_recent_trades(recent_trades_manager):
    # if setting no new trades
    recent_trades_manager.set_all_recent_trades([])
    assert not recent_trades_manager.recent_trades
    recent_trades_manager.set_all_recent_trades(None)
    assert not recent_trades_manager.recent_trades

    # test removing all previous values
    recent_trades_manager.recent_trades.append(1)
    assert 1 in recent_trades_manager.recent_trades
    trades = random_recent_trades()
    recent_trades_manager.set_all_recent_trades(trades)
    assert trades[-1] in recent_trades_manager.recent_trades
    assert 1 not in recent_trades_manager.recent_trades

    # test set values
    recent_trade_1 = random_recent_trade()
    recent_trade_2 = random_recent_trade()
    recent_trade_3 = random_recent_trade()
    trades = [recent_trade_3, recent_trade_2, recent_trade_1]
    recent_trades_manager.set_all_recent_trades(trades)
    assert recent_trade_1 in recent_trades_manager.recent_trades
    assert recent_trade_2 in recent_trades_manager.recent_trades
    assert recent_trades_manager.recent_trades[-1] == recent_trade_1
    assert recent_trades_manager.recent_trades[-2] == recent_trade_2
    assert recent_trades_manager.recent_trades[-3] == recent_trade_3


async def test_add_new_trades(recent_trades_manager):
    # if adding no new trades
    recent_trades_manager.add_new_trades([])
    assert not recent_trades_manager.recent_trades
    recent_trades_manager.add_new_trades(None)
    assert not recent_trades_manager.recent_trades

    # test set values
    recent_trade_1 = random_recent_trade()
    recent_trade_2 = random_recent_trade()
    recent_trade_3 = random_recent_trade()
    recent_trade_4 = random_recent_trade()
    recent_trade_5 = random_recent_trade()
    trades = [recent_trade_1, recent_trade_2]
    recent_trades_manager.add_new_trades(trades)
    assert recent_trade_1 in recent_trades_manager.recent_trades
    assert recent_trade_2 in recent_trades_manager.recent_trades
    assert recent_trades_manager.recent_trades[0] == recent_trade_1
    assert recent_trades_manager.recent_trades[1] == recent_trade_2
    with pytest.raises(IndexError):
        recent_trades_manager.recent_trades[2]

    # test add duplicate
    trades = [recent_trade_2, recent_trade_3, recent_trade_4]
    recent_trades_manager.add_new_trades(trades)
    assert recent_trades_manager.recent_trades[0] == recent_trade_1
    assert recent_trades_manager.recent_trades[1] == recent_trade_2
    assert recent_trades_manager.recent_trades[2] == recent_trade_3
    assert recent_trades_manager.recent_trades[3] == recent_trade_4
    with pytest.raises(IndexError):
        recent_trades_manager.recent_trades[4]

    trades = [recent_trade_2, recent_trade_5]
    recent_trades_manager.add_new_trades(trades)
    assert recent_trades_manager.recent_trades[0] == recent_trade_1
    assert recent_trades_manager.recent_trades[1] == recent_trade_2
    assert recent_trades_manager.recent_trades[2] == recent_trade_3
    assert recent_trades_manager.recent_trades[3] == recent_trade_4
    assert recent_trades_manager.recent_trades[4] == recent_trade_5


async def test_add_new_liquidations(recent_trades_manager):
    # if adding no new liquidations
    recent_trades_manager.add_new_liquidations([])
    assert not recent_trades_manager.liquidations
    recent_trades_manager.add_new_liquidations(None)
    assert not recent_trades_manager.liquidations


async def test_reset_recent_trades(recent_trades_manager):
    if not os.getenv('CYTHON_IGNORE'):
        recent_trades_manager.recent_trades.append(1)
        recent_trades_manager.liquidations.append(2)
        assert recent_trades_manager.recent_trades
        assert recent_trades_manager.liquidations
        recent_trades_manager._reset_recent_trades()
        assert not recent_trades_manager.recent_trades
        assert not recent_trades_manager.liquidations
        with pytest.raises(IndexError):
            recent_trades_manager.recent_trades[0]
        with pytest.raises(IndexError):
            recent_trades_manager.liquidations[0]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt
import mock
import pytest
import asyncio

import octobot_trading.exchanges as exchanges
import octobot_trading.constants as constants
import octobot_trading.errors as errors



# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_retried_failed_network_request():
  request_mock = mock.AsyncMock(return_value=1)
  @exchanges.connectors.util.retried_failed_network_request()
  async def _fetch_request(*args, **kwargs):
    return await request_mock(*args, **kwargs)

  with mock.patch.object(asyncio, "sleep") as sleep_mock:
    # no error
    assert await _fetch_request("arg1", kw1=1) == 1
    request_mock.assert_awaited_once_with("arg1", kw1=1)
    sleep_mock.assert_not_called()

    # non retriable errors
    for raised_error in [
      ccxt.ExchangeError("Exchange error"),
      KeyError("random key error"),
      ValueError("random value error"),
      ZeroDivisionError("random zero division error"),
    ]:
      request_mock = mock.AsyncMock(side_effect=raised_error)
      @exchanges.connectors.util.retried_failed_network_request()
      async def _fetch_request(*args, **kwargs):
        return await request_mock(*args, **kwargs)
      with pytest.raises(raised_error.__class__):
        await _fetch_request("arg1", kw1=1)
      assert request_mock.call_count == 1
      assert sleep_mock.call_count == 0
      sleep_mock.reset_mock()

    # retriable errors
    for raised_error in [
      ccxt.ExchangeNotAvailable("Exchange not available"), 
      ccxt.RequestTimeout("Request timeout"), 
      ccxt.InvalidNonce("Invalid nonce"),
      errors.RetriableFailedRequest("Retriable failed request"),
    ]:
      request_mock = mock.AsyncMock(side_effect=raised_error)
      @exchanges.connectors.util.retried_failed_network_request()
      async def _fetch_request(*args, **kwargs):
        return await request_mock(*args, **kwargs)
      with pytest.raises(raised_error.__class__):
        await _fetch_request("arg1", kw1=1)
      assert request_mock.call_count == constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS
      assert sleep_mock.call_count == constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS - 1
      assert sleep_mock.mock_calls[0].args == (constants.DEFAULT_FAILED_REQUEST_RETRY_TIME,)
      sleep_mock.reset_mock()

    # retriable exchange errors
    for raised_error in [
      ccxt.ExchangeError("Internal Server Error"),
      ccxt.BadRequest("Error: socket hang up"),
      ccxt.BadRequest("Error: read ECONNRESET"),
      ccxt.ExchangeError('mexc {"code":700022,"msg":"Internal Server Error"}'),
    ]:
      request_mock = mock.AsyncMock(side_effect=raised_error)
      @exchanges.connectors.util.retried_failed_network_request()
      async def _fetch_request(*args, **kwargs):
        return await request_mock(*args, **kwargs)
      with pytest.raises(raised_error.__class__):
        await _fetch_request("arg1", kw1=1)
      assert request_mock.call_count == constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS
      assert sleep_mock.call_count == constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS - 1
      assert sleep_mock.mock_calls[0].args == (constants.DEFAULT_FAILED_REQUEST_RETRY_TIME,)
      sleep_mock.reset_mock()

    # retriable proxy errors
    for raised_error in [
      errors.RetriableExchangeProxyError("Retriable exchange proxy error"),
    ]:
      request_mock = mock.AsyncMock(side_effect=raised_error)
      @exchanges.connectors.util.retried_failed_network_request()
      async def _fetch_request(*args, **kwargs):
        return await request_mock(*args, **kwargs)
      with pytest.raises(raised_error.__class__):
        await _fetch_request("arg1", kw1=1)
      assert request_mock.call_count == constants.FAILED_PROXY_NETWORK_REQUEST_RETRY_ATTEMPTS
      assert sleep_mock.call_count == constants.FAILED_PROXY_NETWORK_REQUEST_RETRY_ATTEMPTS - 1
      assert sleep_mock.mock_calls[0].args == (constants.DEFAULT_FAILED_REQUEST_RETRY_TIME,)
      sleep_mock.reset_mock()

    # custom params
    request_mock = mock.AsyncMock(side_effect=ccxt.ExchangeNotAvailable("Exchange not available"))
    @exchanges.connectors.util.retried_failed_network_request(
      attempts=3, retriable_proxy_errors_attempts=1, delay=0.1
    )
    async def _fetch_request(*args, **kwargs):
      return await request_mock(*args, **kwargs)
    with pytest.raises(ccxt.ExchangeNotAvailable):
      await _fetch_request("arg1", kw1=1)
    assert request_mock.call_count == 3
    assert sleep_mock.call_count == 2
    assert sleep_mock.mock_calls[0].args == (0.1,)
    sleep_mock.reset_mock()

    request_mock = mock.AsyncMock(side_effect=errors.RetriableExchangeProxyError("Retriable exchange proxy error"))
    @exchanges.connectors.util.retried_failed_network_request(
      attempts=4, retriable_proxy_errors_attempts=4, delay=0.2
    )
    async def _fetch_request(*args, **kwargs):
      return await request_mock(*args, **kwargs)
    with pytest.raises(errors.RetriableExchangeProxyError):
      await _fetch_request("arg1", kw1=1)
    assert request_mock.call_count == 4
    assert sleep_mock.call_count == 3
    assert sleep_mock.mock_calls[0].args == (0.2,)
    sleep_mock.reset_mock()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.transactions.transaction as transaction


class BlockchainTransaction(transaction.Transaction):
    def __init__(self, exchange_name, creation_time, transaction_type, currency, blockchain_type,
                 blockchain_transaction_id,
                 blockchain_transaction_status=enums.BlockchainTransactionStatus.CREATED,
                 source_address=None,
                 destination_address=None,
                 quantity=constants.ZERO,
                 transaction_fee=constants.ZERO):
        self.source_address = source_address
        self.destination_address = destination_address
        self.blockchain_transaction_id = blockchain_transaction_id
        self.blockchain_type = blockchain_type
        self.blockchain_transaction_status = blockchain_transaction_status
        self.quantity = quantity
        self.transaction_fee = transaction_fee
        super().__init__(exchange_name, creation_time, transaction_type, currency=currency)
        self.transaction_id = self.blockchain_transaction_id if self.blockchain_transaction_id else str(uuid.uuid4())

    def is_deposit(self):
        return self.transaction_type is enums.TransactionType.BLOCKCHAIN_DEPOSIT

    def is_withdrawal(self):
        return self.transaction_type is enums.TransactionType.BLOCKCHAIN_WITHDRAWAL

    def is_pending(self):
        return self.blockchain_transaction_status is enums.BlockchainTransactionStatus.CONFIRMING

    def is_validated(self):
        return self.blockchain_transaction_status is enums.BlockchainTransactionStatus.SUCCESS



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums
import octobot_trading.errors as errors


POSITION_DICT_DECIMAL_KEYS = [
    enums.ExchangeConstantsPositionColumns.QUANTITY.value,
    enums.ExchangeConstantsPositionColumns.SIZE.value,
    enums.ExchangeConstantsPositionColumns.NOTIONAL.value,
    enums.ExchangeConstantsPositionColumns.INITIAL_MARGIN.value,
    enums.ExchangeConstantsPositionColumns.COLLATERAL.value,
    enums.ExchangeConstantsPositionColumns.LEVERAGE.value,
    enums.ExchangeConstantsPositionColumns.ENTRY_PRICE.value,
    enums.ExchangeConstantsPositionColumns.MARK_PRICE.value,
    enums.ExchangeConstantsPositionColumns.LIQUIDATION_PRICE.value,
    enums.ExchangeConstantsPositionColumns.UNREALIZED_PNL.value,
    enums.ExchangeConstantsPositionColumns.REALISED_PNL.value,
    enums.ExchangeConstantsPositionColumns.MAINTENANCE_MARGIN_RATE.value,
]


def create_position_instance_from_raw(trader, raw_position):
    """
    Creates a position instance from a raw position dictionary
    :param trader: the trader instance
    :param raw_position: the raw position dictionary
    :return: the created position
    """
    position_symbol_contract = trader.exchange_manager.exchange.get_pair_future_contract(
        raw_position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value)
    )
    position = create_position_from_type(trader, position_symbol_contract)
    position.update_from_raw(raw_position)
    return position


def create_position_instance_from_dict(trader, position_dict: dict):
    """
    Creates a position instance from a raw position dictionary
    :param trader: the trader instance
    :param position_dict: the raw position dictionary as from position.to_dict
    :return: the created position
    """
    raw_position = sanitize_raw_decimals(position_dict)
    return create_position_instance_from_raw(trader, raw_position)


def sanitize_raw_position(position_dict: dict):
    position_dict[enums.ExchangeConstantsPositionColumns.STATUS.value] = (
        personal_data.parse_position_status(position_dict))
    position_dict[enums.ExchangeConstantsPositionColumns.SIDE.value] = (
        personal_data.parse_position_side(position_dict))
    position_dict[enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value] = (
        personal_data.parse_position_margin_type(position_dict))
    position_dict[enums.ExchangeConstantsPositionColumns.POSITION_MODE.value] = (
        personal_data.parse_position_mode(position_dict))
    return sanitize_raw_decimals(position_dict)


def sanitize_raw_decimals(position_dict: dict):
    for key in POSITION_DICT_DECIMAL_KEYS:
        value = position_dict.get(key)
        if value is not None and value != "":
            position_dict[key] = decimal.Decimal(str(value))
    return position_dict


def create_position_from_type(trader, symbol_contract):
    """
    Creates a position instance from a position type
    :param trader: the trader instance
    :param symbol_contract: the position symbol contract
    :return: the created position
    """
    if symbol_contract.is_handled_contract():
        return personal_data.TraderPositionTypeClasses[symbol_contract.contract_type](trader, symbol_contract)
    raise errors.UnhandledContractError(f"{symbol_contract} is not supported")


def create_symbol_position(trader, symbol):
    """
    Creates a position for a specified symbol
    :param trader: the trader instance
    :param symbol: the position symbol
    :return: the created position
    """
    return create_position_instance_from_raw(trader, {
        enums.ExchangeConstantsPositionColumns.SYMBOL.value: symbol
    })



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing
import contextlib
import decimal
import copy

import octobot_commons.logging as logging

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.states as orders_states
import octobot_trading.exchanges
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.trailing_profiles as trailing_profiles
import octobot_trading.personal_data.orders.order_group as order_group_import
import octobot_trading.personal_data.orders.order_state as order_state_import
import octobot_trading.personal_data.orders.decimal_order_adapter as decimal_order_adapter
import octobot_trading.personal_data.orders.triggers.base_trigger as base_trigger_import
import octobot_trading.personal_data.orders.cancel_policies.order_cancel_policy as order_cancel_policy_import
import octobot_trading.personal_data.orders.cancel_policies.cancel_policy_factory as cancel_policy_factory
import octobot_trading.util as util


class Order(util.Initializable):
    """
    Order class will represent an open order in the specified exchange
    In simulation it will also define rules to be filled / canceled
    It is also use to store creation & fill values of the order
    """
    CHECK_ORDER_STATUS_AFTER_INIT_DELAY = 2
    SUPPORTS_GROUPING = True    # False when orders of this type can't be grouped
    USE_ORIGIN_QUANTITY_AS_FILLED_QUANTITY = False

    def __init__(self, trader, side=None):
        super().__init__()
        self.trader: octobot_trading.exchanges.Trader = trader
        self.exchange_manager: octobot_trading.exchanges.ExchangeManager = trader.exchange_manager
        self.lock: asyncio.Lock = asyncio.Lock()
        self.is_synchronized_with_exchange: bool = False
        self.is_from_this_octobot: bool = True
        self.simulated: bool = trader.simulate

        self.logger_name: typing.Optional[str] = None
        self.order_id: str = order_util.generate_order_id()        # used id; kept through instances and trading signals
        self.exchange_order_id: str = trader.parse_order_id(None)  # given by the exchange, local to the user account
        self.status: enums.OrderStatus = enums.OrderStatus.OPEN
        self.symbol: str = None # type: ignore
        self.currency: typing.Optional[str] = None
        self.market: typing.Optional[str] = None
        self.quantity_currency: typing.Optional[str] = None
        self.taker_or_maker: typing.Optional[str] = None
        self.timestamp: float = 0
        self.side: enums.TradeOrderSide = side # type: ignore
        self.trigger_above: bool = None # type: ignore
        self.tag: str = None # type: ignore
        self.associated_entry_ids: typing.Optional[list[str]] = None
        self.broker_applied: bool = False

        # original order attributes
        self.creation_time: float = self.exchange_manager.exchange.get_exchange_current_time()
        self.origin_price: decimal.Decimal = constants.ZERO
        self.created_last_price: decimal.Decimal = constants.ZERO
        self.origin_quantity: decimal.Decimal = constants.ZERO
        self.origin_stop_price: decimal.Decimal = constants.ZERO

        # order type attributes
        self.order_type: enums.TraderOrderType = None # type: ignore
        # raw exchange order type, used to create order dict
        self.exchange_order_type: typing.Optional[enums.TradeOrderType] = None

        # filled order attributes
        self.filled_quantity: decimal.Decimal = constants.ZERO
        self.filled_price: decimal.Decimal = constants.ZERO
        self.fee: typing.Optional[dict[str, typing.Any]] = None
        self.fees_currency_side: enums.FeesCurrencySide = enums.FeesCurrencySide.UNDEFINED
        self.total_cost: decimal.Decimal = constants.ZERO
        self.order_profitability: decimal.Decimal = constants.ZERO
        self.executed_time: float = 0

        # canceled order attributes
        self.canceled_time: float = 0

        self.order_group: typing.Optional[order_group_import.OrderGroup] = None
        self.trailing_profile: typing.Optional[trailing_profiles.TrailingProfile] = None

        # order state is initialized in initialize_impl()
        self.state: typing.Optional[order_state_import.OrderState] = None

        # order activity
        self.is_active: bool = True   # When is_active=False order is not pushed to exchanges
        # True when a transition between active and inactive is being made
        self.is_in_active_inactive_transition: bool = False
        # active_trigger is used for active/inactive switch trigger mechanism, it stores relevant data.
        self.active_trigger: typing.Optional[base_trigger_import.BaseTrigger] = None

        # future trading attributes
        # when True: reduce position quantity only without opening a new position if order.quantity > position.quantity
        self.reduce_only: bool = False

        # when True: close the current associated position
        self.close_position: bool = False

        # the associated position side (should be BOTH for One-way Mode ; LONG or SHORT for Hedge Mode)
        self.position_side: typing.Optional[enums.PositionSide] = None

        # Chained orders attributes
        # other orders (as any Order) that should be created when this order is filled
        self.chained_orders: list[Order] = []
        # order that triggered this order creation (when created as a chained order)
        self.triggered_by: typing.Optional[Order] = None
        # if True this orders quantity will be reduced according to the triggering order's paid fees
        self.update_with_triggering_order_fees: bool = False
        # True when this order has been created directly by the exchange (usually as stop loss / take profit
        # when passed as parameter alongside another order)
        self.has_been_bundled: bool = False
        # True when this order is to be opened as a chained order and has not been open yet
        self.is_waiting_for_chained_trigger: bool = False
        # instance of the order that has been created after self has been filled. Used in stop / TP orders
        self.on_filled_artificial_order: typing.Optional[Order] = None

        # Cancel policy, if set
        self.cancel_policy: typing.Optional[order_cancel_policy_import.OrderCancelPolicy] = None

        # Params given to the exchange request when this order is created. Include any exchange specific param here.
        # All params and values in those will be ignored in simulated orders
        self.exchange_creation_params: dict[str, typing.Any] = {}
        # kwargs given to trader.create_order() when this order should be created later on
        self.trader_creation_kwargs: dict[str, typing.Any] = {}

    @classmethod
    def get_name(cls):
        return cls.__name__

    def get_logger_name(self):
        if self.logger_name is None:
            self.logger_name = f"{self.get_name()} | {self.order_id} [exchange id: {self.exchange_order_id}]"
        return self.logger_name

    def update(
        self, symbol=None, order_id="", exchange_order_id=None, status=enums.OrderStatus.OPEN,
        current_price=constants.ZERO, quantity=constants.ZERO, price=constants.ZERO, stop_price=constants.ZERO,
        quantity_filled=constants.ZERO, filled_price=constants.ZERO, average_price=constants.ZERO,
        fee=None, total_cost=constants.ZERO, timestamp=None,
        order_type=None, reduce_only=None, close_position=None, position_side=None, fees_currency_side=None,
        group=None, tag=None, quantity_currency=None, exchange_creation_params=None,
        associated_entry_id=None, trigger_above=None, trailing_profile: trailing_profiles.TrailingProfile=None,
        is_active=None, active_trigger: base_trigger_import.BaseTrigger = None,
        cancel_policy: typing.Optional[order_cancel_policy_import.OrderCancelPolicy] = None,
    ) -> bool:
        changed: bool = False
        should_update_total_cost = False

        price = current_price if (current_price and self.use_current_price_as_origin_price()) else price

        if order_id and self.order_id != order_id:
            self.order_id = order_id

        if exchange_order_id and self.exchange_order_id != exchange_order_id:
            self.exchange_order_id = exchange_order_id

        if symbol and self.symbol != symbol:
            self.currency, self.market = self.exchange_manager.get_exchange_quote_and_base(symbol)
            self.symbol = symbol

        if quantity_currency is None:
            if self.quantity_currency is None and self.symbol is not None:
                self.quantity_currency = order_util.get_order_quantity_currency(self.exchange_manager, self.symbol)
        else:
            self.quantity_currency = quantity_currency

        if status and self.status != status:
            # ensure the order status is compatible with the state to avoid exchange sync issues
            if self.state is None or self.state.allows_new_status(status):
                self.status = status
                changed = True
            else:
                logging.get_logger(self.get_logger_name()).debug(f"Ignored unexpected new status: {status}")
        if not self.status:
            self.status = enums.OrderStatus.OPEN

        if timestamp and self.timestamp != timestamp:
            self.timestamp = timestamp
            # if we have a timestamp, it's a real trader => need to format timestamp if necessary
            self.creation_time = self.exchange_manager.exchange.get_uniformized_timestamp(timestamp)
        if not self.timestamp:
            if not timestamp:
                self.creation_time = self.exchange_manager.exchange.get_exchange_current_time()
            else:
                # if we have a timestamp, it's a real trader => need to format timestamp if necessary
                self.creation_time = self.exchange_manager.exchange.get_uniformized_timestamp(timestamp)
            self.timestamp = self.creation_time

        if status in {enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED} and not self.executed_time:
            self.executed_time = self.timestamp

        if price and self.origin_price != price:
            previous_price = self.origin_price
            self.origin_price = price
            self._on_origin_price_change(
                previous_price, self.exchange_manager.exchange.get_exchange_current_time()
            )
            changed = True
            should_update_total_cost = True

        if fee is not None and self.fee != fee:
            self.fee = fee

        if fees_currency_side is not None and self.fees_currency_side != fees_currency_side:
            self.fees_currency_side = fees_currency_side

        if current_price and self.created_last_price != current_price:
            self.created_last_price = current_price
            changed = True

        if quantity and self.origin_quantity != quantity:
            self.origin_quantity = quantity
            changed = True
            should_update_total_cost = True

        if stop_price and self.origin_stop_price != stop_price:
            self.origin_stop_price = stop_price
            changed = True

        if total_cost and self.total_cost != total_cost:
            self.total_cost = total_cost

        if average_price:
            # try using average price first
            if self.filled_price != average_price:
                self.filled_price = average_price
                should_update_total_cost = True
        else:
            if filled_price and self.filled_price != filled_price:
                self.filled_price = filled_price
                should_update_total_cost = True

        if self.trader.simulate:
            if quantity and (self.is_filled() or self.USE_ORIGIN_QUANTITY_AS_FILLED_QUANTITY):
                # when simulating, only set filled quantity if the order is filled
                self.filled_quantity = quantity
                changed = True
                should_update_total_cost = True

            if quantity_filled and self.filled_quantity != quantity_filled:
                self.filled_quantity = quantity_filled
                changed = True
                should_update_total_cost = True
        else:
            if quantity_filled and self.filled_quantity != quantity_filled:
                self.filled_quantity = quantity_filled
                changed = True
                should_update_total_cost = True

        if order_type:
            self.order_type = order_type
            if self.exchange_order_type is None:
                self.exchange_order_type = _get_trade_order_type(order_type)

        if not self.filled_price and self.filled_quantity and self.total_cost:
            self.filled_price = self.total_cost / self.filled_quantity
            if timestamp is not None:
                self.executed_time = self.exchange_manager.exchange.get_uniformized_timestamp(timestamp)

        if self.taker_or_maker is None:
            self._update_taker_maker()

        if position_side:
            self.position_side = position_side

        if reduce_only is not None:
            self.reduce_only = reduce_only

        if close_position is not None:
            self.close_position = close_position

        if group is not None:
            self.add_to_order_group(group)

        if tag is not None:
            self.tag = tag

        if exchange_creation_params is not None:
            self.exchange_creation_params = exchange_creation_params

        if associated_entry_id is not None:
            self.associate_to_entry(associated_entry_id)

        if trigger_above is not None and self.trigger_above != trigger_above:
            changed = True
            self.trigger_above = trigger_above

        if trailing_profile is not None and self.trailing_profile != trailing_profile:
            changed = True
            self.trailing_profile = trailing_profile

        if is_active is not None and self.is_active != is_active:
            changed = True
            self.is_active = is_active

        if active_trigger is not None and active_trigger != self.active_trigger:
            changed = True
            self.use_active_trigger(active_trigger)

        if cancel_policy is not None and cancel_policy != self.cancel_policy:
            changed = True
            self.cancel_policy = cancel_policy

        if should_update_total_cost and not total_cost:
            self._update_total_cost()

        return changed

    async def create_on_filled_artificial_order(self, enable_associated_orders_creation):
        """
        :return: the equivalent artificial order after trigger price (ex: stop loss => market order)
        """
        await self.on_filled(enable_associated_orders_creation, force_artificial_orders=True)
        if self.on_filled_artificial_order is not None:
            logging.get_logger(self.logger_name).info(f"Created artificial order: {self.on_filled_artificial_order}")
            if self.order_group:
                self.on_filled_artificial_order.add_to_order_group(self.order_group)
        return self.on_filled_artificial_order

    async def initialize_impl(self, **kwargs):
        """
        Initialize order status update tasks
        """
        await orders_states.create_order_state(self, **kwargs)
        if self.is_created() and not self.is_closed():
            await self.update_order_status()
        if not self.is_active:
            await self._ensure_inactive_order_watcher()

    def register_broker_applied_if_enabled(self):
        if not self.simulated and self.trader and self.trader.exchange_manager:
            self.broker_applied = self.trader.exchange_manager.is_broker_enabled

    def _on_origin_price_change(self, previous_price, price_time):
        """
        Called when origin price just changed.
        Override if necessary
        :param previous_price: the previous origin_price
        :param price_time: time starting from when the price should be considered
        """
        if self.order_group and self.order_group.active_order_swap_strategy:
            self.order_group.active_order_swap_strategy.on_order_update(self, price_time)

    def add_chained_order(self, chained_order):
        """
        chained_order will be assigned with the actually created order when this order will be filled
        warning: add_chained_order is not checking if the order should be created instantly (if self is filled).
        :param chained_order: WrappedOrder to be added to this order's chained orders
        """
        self.chained_orders.append(chained_order)

    async def update_order_status(self, force_refresh=False):
        """
        Update_order_status will define the rules for a simulated order to be filled / canceled
        """
        raise NotImplementedError("Update_order_status not implemented")

    def add_to_order_group(self, order_group):
        if not self.is_open() and not self.is_waiting_for_chained_trigger:
            logging.get_logger(self.get_logger_name()).warning(f"Adding order to group however order is not open.")
        self.order_group = order_group

    def get_total_fees(self, currency):
        return order_util.get_fees_for_currency(self.fee, currency)

    def is_created(self) -> bool:
        return self.state is None or self.state.is_created()

    def is_pending_creation(self) -> bool:
        return isinstance(self.state, orders_states.PendingCreationOrderState)

    def is_open(self) -> bool:
        # also check is_initialized to avoid considering uncreated orders as open
        return (
            self.state is None or self.state.is_open() or (
                not self.is_active and self.status is enums.OrderStatus.OPEN and not self.is_waiting_for_chained_trigger
            )
        )

    def is_filled(self) -> bool:
        if self.state is None:
            return self.status is enums.OrderStatus.FILLED
        return self.state.is_filled() or (self.state.is_closed() and self.status is enums.OrderStatus.FILLED)

    def is_cancelled(self) -> bool:
        if self.state is None:
            return self.status is enums.OrderStatus.CANCELED
        return self.state.is_canceled() or (self.state.is_closed() and self.status is enums.OrderStatus.CANCELED)

    def is_cancelling(self) -> bool:
        if self.state is None:
            return self.status is enums.OrderStatus.PENDING_CANCEL
        return self.state.state is enums.OrderStates.CANCELING or self.status is enums.OrderStatus.PENDING_CANCEL

    def is_closed(self) -> bool:
        if self.state is None:
            return self.status is enums.OrderStatus.CLOSED
        return self.state.is_closed() if self.state is not None else self.status is enums.OrderStatus.CLOSED

    def is_refreshing(self) -> bool:
        return self.state is not None and self.state.is_refreshing()

    def is_pending(self) -> bool:
        return self.state is not None and self.state.is_pending()

    def is_refreshing_filling_state(self) -> bool:
        return self._is_refreshing_state(orders_states.FillOrderState)

    def is_refreshing_canceling_state(self) -> bool:
        return self._is_refreshing_state(orders_states.CancelOrderState)

    def is_pending_cancel_state(self) -> bool:
        return self._is_pending_state(orders_states.CancelOrderState)

    def _is_refreshing_state(self, state_type) -> bool:
        return self.is_refreshing() and isinstance(self.state, state_type)

    def _is_pending_state(self, state_type) -> bool:
        return self.is_pending() and isinstance(self.state, state_type)

    def can_be_edited(self) -> bool:
        # orders that are not yet open or already open can be edited
        return self.state is None or (self.state.is_open() and not self.is_refreshing())

    def use_current_price_as_origin_price(self):
        # Override to return True when the current order price can't be set by the user (ex: market orders)
        return False

    def get_position_side(self, future_contract):
        """
        :param future_contract: the associated future contract
        :return: the position side if it can be determined, else raise an InvalidPositionSide
        """
        if self.position_side is not None:
            return self.position_side
        if future_contract.is_one_way_position_mode():
            return enums.PositionSide.BOTH
        raise errors.InvalidPositionSide(f"Can't determine order position side while using hedge position mode")

    @contextlib.contextmanager
    def active_or_inactive_transition(self):
        previous_value = self.is_in_active_inactive_transition
        try:
            self.is_in_active_inactive_transition = True
            yield
        finally:
            self.is_in_active_inactive_transition = previous_value

    def use_active_trigger(self, active_trigger: base_trigger_import.BaseTrigger):
        if active_trigger is None:
            raise ValueError("active_trigger must be provided")
        if self.active_trigger is None:
            self.active_trigger = active_trigger
        elif self.active_trigger.is_pending():
            logging.get_logger(self.get_logger_name()).error(
                f"The current active trigger ({str(self.active_trigger)}) is still pending, canceling it "
                f"and replacing it by this new one, this is works but is unexpected."
            )
            self.active_trigger.clear()
            self.active_trigger = active_trigger
        else:
            self.active_trigger.update_from_other_trigger(active_trigger)

    async def set_as_inactive(self, active_trigger: base_trigger_import.BaseTrigger):
        """
        Marks the instance as inactive and ensures the inactive order watcher is scheduled.
        """
        logging.get_logger(self.get_logger_name()).info("Order is switching to inactive")
        self.use_active_trigger(active_trigger)
        self.is_active = False
        # enforce attributes in case order has been canceled
        self.status = enums.OrderStatus.OPEN
        self.canceled_time = 0
        await self._ensure_inactive_order_watcher()

    def should_become_active(self, price_time: float, current_price: decimal.Decimal) -> bool:
        if self.is_active:
            return False
        if price_time >= self.creation_time:
            return self.active_trigger.triggers(current_price)
        return False

    async def _ensure_inactive_order_watcher(self):
        if self.is_active:
            # order is active, nothing to do
            return
        if self.is_waiting_for_chained_trigger:
            # watcher should be created only when this chained order is created
            # will be called again once the chained order gets created
            return
        if not self.exchange_manager.trader.enable_inactive_orders or self.is_self_managed():
            # can't be inactive
            logging.get_logger(self.get_logger_name()).error(
                f"Unexpected inactive order (simulated={self.simulated} self_managed={self.is_self_managed()}): {self}"
            )
            return
        if self.active_trigger is None:
            logging.get_logger(self.get_logger_name()).error("self.active_trigger is None")
            return
        if self.is_synchronization_enabled():
            logging.get_logger(self.get_logger_name()).debug(
                f"Creating watcher for inactive order: {str(self)} - trigger: {self.active_trigger}"
            )
            await self.active_trigger.create_watcher(self.exchange_manager, self.symbol, self.creation_time)

    @contextlib.contextmanager
    def order_state_creation(self):
        try:
            yield
        except errors.InvalidOrderState as exc:
            logging.get_logger(self.get_logger_name()).exception(exc, True, f"Error when creating order state: {exc}")

    async def on_inactive_from_active(self):
        """
        Update the order to be considered as "confirmed" inactive. Called when the order was active before
        """
        if self.active_trigger is None:
            raise ValueError(
                f"self.active_trigger must be provided to set an order as inactive"
            )
        await self.set_as_inactive(self.active_trigger)
        self.clear_active_order_elements()

    async def on_active_from_inactive(self):
        """
        Update the order to be considered as "confirmed" active => the new active order was created
        Self has already been removed from order manager by the new active order as they share the same order_id
        """
        self.is_active = True
        self.clear()

    async def on_active_trigger(
        self, strategy_timeout: typing.Optional[float], wait_for_fill_callback: typing.Optional[typing.Callable]
    ):
        try:
            if self.is_active:
                logging.get_logger(self.get_logger_name()).error("Skipped active trigger for an already active order.")
                return
            logging.get_logger(self.get_logger_name()).info("Order is becoming active.")
            await order_util.create_as_active_order_using_strategy_if_any(self, strategy_timeout, wait_for_fill_callback)
        except Exception as err:
            logging.get_logger(self.get_logger_name()).exception(
                err,
                True,
                f"Unexpected error ({err.__class__.__name__}: {err}) "
                f"when creating active order from inactive order: {self}"
            )

    async def on_pending_creation(self, is_from_exchange_data=False, enable_associated_orders_creation=True,
        is_already_counted_in_available_funds=False
    ):
        with self.order_state_creation():
            state_class = orders_states.PendingCreationChainedOrderState if self.is_waiting_for_chained_trigger \
                else orders_states.PendingCreationOrderState
            self.state = state_class(
                self, is_from_exchange_data=is_from_exchange_data,
                enable_associated_orders_creation=enable_associated_orders_creation,
                is_already_counted_in_available_funds=is_already_counted_in_available_funds
            )
            await self.state.initialize()

    async def on_open(
        self, force_open=False, is_from_exchange_data=False, enable_associated_orders_creation=True,
        is_already_counted_in_available_funds=False
    ):
        with self.order_state_creation():
            if isinstance(self.state, orders_states.PendingCreationOrderState):
                await self.state.trigger_terminate()
            if isinstance(self.state, orders_states.OpenOrderState):
                if not self.state.is_initialized:
                    logging.get_logger(self.get_logger_name()).error(f"on_open called with existing "
                                                                     f"uninitialized OpenOrderState.")
                # state has already been created and initialized
                return
            self.state = orders_states.OpenOrderState(
                self, is_from_exchange_data=is_from_exchange_data,
                enable_associated_orders_creation=enable_associated_orders_creation,
                is_already_counted_in_available_funds=is_already_counted_in_available_funds
            )
            await self.state.initialize(forced=force_open)

    async def on_fill(
        self, force_fill=False, is_from_exchange_data=False, enable_associated_orders_creation=None,
        is_already_counted_in_available_funds=False
    ):
        enable_associated_orders_creation = self.state.enable_associated_orders_creation \
            if (self.state and enable_associated_orders_creation is None) \
            else (True if enable_associated_orders_creation is None else enable_associated_orders_creation)
        if self.is_in_active_inactive_transition:
            logging.get_logger(self.get_logger_name()).info("Completing active-inactive transition: order is filled")
            self.is_in_active_inactive_transition = False
        if (self.is_open() and not self.is_refreshing()) or self.is_pending_creation():
            with self.order_state_creation():
                self.state = orders_states.FillOrderState(
                    self, is_from_exchange_data=is_from_exchange_data,
                    enable_associated_orders_creation=enable_associated_orders_creation,
                    is_already_counted_in_available_funds=is_already_counted_in_available_funds
                )
                await self.state.initialize(forced=force_fill)
        else:
            logging.get_logger(self.get_logger_name()).debug(f"Trying to fill a refreshing or previously filled "
                                                             f"or canceled order: "
                                                             f"ignored fill call for {self}")

    async def on_close(self, force_close=False, is_from_exchange_data=False, enable_associated_orders_creation=None,
        is_already_counted_in_available_funds=False
    ):
        enable_associated_orders_creation = self.state.enable_associated_orders_creation \
            if (self.state and enable_associated_orders_creation is None) \
            else (True if enable_associated_orders_creation is None else enable_associated_orders_creation)
        with self.order_state_creation():
            self.state = orders_states.CloseOrderState(
                self, is_from_exchange_data=is_from_exchange_data,
                enable_associated_orders_creation=enable_associated_orders_creation,
                is_already_counted_in_available_funds=is_already_counted_in_available_funds
            )
            await self.state.initialize(forced=force_close)

    async def on_cancel(
        self, is_from_exchange_data=False, force_cancel=False, enable_associated_orders_creation=None,
        ignored_order=None, is_already_counted_in_available_funds=False
    ):
        enable_associated_orders_creation = self.state.enable_associated_orders_creation \
            if (self.state and enable_associated_orders_creation is None) \
            else (True if enable_associated_orders_creation is None else enable_associated_orders_creation)
        with self.order_state_creation():
            self.state = orders_states.CancelOrderState(
                self, is_from_exchange_data=is_from_exchange_data,
                enable_associated_orders_creation=enable_associated_orders_creation,
                is_already_counted_in_available_funds=is_already_counted_in_available_funds
            )
            await self.state.initialize(forced=force_cancel, ignored_order=ignored_order)

    def on_fill_actions(self):
        """
        Perform on_fill actions
        """
        self.status = enums.OrderStatus.FILLED

    async def on_filled(self, enable_associated_orders_creation, force_artificial_orders=False):
        """
        Filling complete callback
        """
        if enable_associated_orders_creation:
            await self._trigger_chained_orders(enable_associated_orders_creation)
        elif self.chained_orders:
            logging.get_logger(self.get_logger_name()).info(
                f"Skipped chained orders creation for {len(self.chained_orders)} chained orders: "
                f"enable_associated_orders_creation is {enable_associated_orders_creation}"
            )

    def associate_to_entry(self, entry_order_id):
        if self.associated_entry_ids is None:
            self.associated_entry_ids = []
        if entry_order_id not in self.associated_entry_ids:
            self.associated_entry_ids.append(entry_order_id)
            return True
        return False

    def update_quantity_with_order_fees(self, other_order_or_trade):
        relevant_fees_amount = order_util.get_fees_for_currency(other_order_or_trade.fee, self.quantity_currency)
        if relevant_fees_amount:
            logger = logging.get_logger(self.get_logger_name())
            fees_str = f"Paid {self.quantity_currency} fees: {relevant_fees_amount}, " \
                       f"initial order size: {self.origin_quantity}"
            if relevant_fees_amount >= self.origin_quantity:
                logger.error(f"Impossible to update chained order amount according to triggering order fees: "
                             f"fees are larger than then chained order size. {fees_str}")
                return False
            self.origin_quantity = decimal_order_adapter.decimal_adapt_quantity(
                self.exchange_manager.exchange.get_market_status(self.symbol, with_fixer=False),
                self.origin_quantity - relevant_fees_amount
            )
            fees_str = f"{fees_str}, updated size: {self.origin_quantity}"
            logger.debug(f"Updating chained order quantity with triggering order fees. {fees_str}")
        return True

    async def update_price_if_outdated(self):
        """
        Implement if necessary
        """

    async def _trigger_chained_orders(self, enable_associated_orders_creation):
        logger = logging.get_logger(self.get_logger_name())
        for index, order in enumerate(self.chained_orders):
            if order.is_cleared():
                logger.error(
                    f"Chained order {index + 1}/{len(self.chained_orders)} has been cleared: skipping "
                    f"creation (order: {order})"
                )
                continue
            can_be_created = await order_util.adapt_chained_order_before_creation(self, order)
            if can_be_created and order.should_be_created():
                logger.info(f"Creating chained order {index + 1}/{len(self.chained_orders)}")
                await self._create_triggered_chained_order(order, enable_associated_orders_creation)
            else:
                logger.info(f"Skipping cancelled chained order {index + 1}/{len(self.chained_orders)}")

    async def _create_triggered_chained_order(self, order, enable_associated_orders_creation):
        logger = logging.get_logger(self.get_logger_name())
        try:
            await order_util.create_as_chained_order(order)
        except errors.ExchangeClosedPositionError as err:
            message = (
                f"chained order is cancelled as it can't be created: position is closed "
                f"on exchange ({err}) order: {order.order_id}"
            )
            if order.reduce_only:
                # can happen on reduce only orders
                logger.warning(f"Reduce only {message}")
            else:
                # should not happen on non-reduce only orders
                logger.error(f"Unexpected: Non reduce only {message}")
            # order can't be created: consider it cancelled
            # note: grouped orders will also be skipped as this one is cancelled (should_be_created will be false)
            order.status = enums.OrderStatus.CLOSED
            # clear state to avoid using outdated pending creation state
            order.state = None
            order.clear()
        except errors.ExchangeOrderInstantTriggerError as err:
            logger.info(f"Order would instantly trigger ({err}). Creating artificial order instead.")
            equivalent_order = await order.create_on_filled_artificial_order(enable_associated_orders_creation)
            if equivalent_order is None:
                # should not happen on limit or market only orders
                logger.error(f"Unexpected instant trigger error: {err} when creating chained order {order}")
            else:
                # acceptable: convert this order into its "triggered artificial order" equivalent
                logger.warning(
                    f"Outdated chained order: order trigger price has already been reached. "
                    f"Creating equivalent {equivalent_order.get_name()} artificial order instead. "
                    f"Initial order: {order}, artificial order: {equivalent_order}."
                )
            # note: grouped orders will also be skipped as this one is filled (should_be_created will be false)
            order.status = enums.OrderStatus.CLOSED
            # clear state to avoid using outdated pending creation state
            order.state = None
            order.clear()
        except Exception as err:
            logger.exception(
                err,
                True,
                f"Unexpected error ({err.__class__.__name__}: {err}) when creating chained order: {order}"
            )

    async def set_as_chained_order(self, triggered_by, has_been_bundled, exchange_creation_params,
                                   update_with_triggering_order_fees, **trader_creation_kwargs):
        if triggered_by is self:
            raise errors.ConflictingOrdersError("Impossible to chain an order to itself")
        self.triggered_by = triggered_by
        # update forecasted fees as initial price will now be the triggered_by filling price
        self._update_taker_maker()
        self.update_with_triggering_order_fees = update_with_triggering_order_fees
        self.has_been_bundled = has_been_bundled
        self.exchange_creation_params = exchange_creation_params
        self.trader_creation_kwargs = trader_creation_kwargs
        self.is_waiting_for_chained_trigger = True
        self.status = enums.OrderStatus.PENDING_CREATION
        await self.initialize()

    def should_be_created(self):
        return not self.is_created() and self.is_waiting_for_chained_trigger and \
            not self._are_simultaneously_triggered_grouped_orders_closed()

    def _are_simultaneously_triggered_grouped_orders_closed(self):
        if self.triggered_by is None:
            return False
        for other_order in self.triggered_by.chained_orders:
            if other_order is self:
                continue
            order_to_check = (
                other_order
                if other_order.on_filled_artificial_order is None
                else other_order.on_filled_artificial_order
            )
            if (
                self.order_group is not None and self.order_group is order_to_check.order_group
                and (
                    order_to_check.is_closed()  # grouped order has been closed
                    or not order_to_check.SUPPORTS_GROUPING # grouped order can't be used in groups (anymore)
                )
            ):
                return True
        return False

    def has_exchange_fetched_fees(self):
        if not self.fee:
            return False
        try:
            # requires fees to be from exchange and having a not None exchange original cost
            return self.fee[enums.FeePropertyColumns.IS_FROM_EXCHANGE.value] \
                   and self.fee[enums.FeePropertyColumns.EXCHANGE_ORIGINAL_COST.value] is not None
        except KeyError:
            return False

    def get_computed_fee(self, forced_value=None, use_origin_quantity_and_price=False):
        is_from_exchange = False
        price = self.origin_price if use_origin_quantity_and_price else self.filled_price
        quantity = self.origin_quantity if use_origin_quantity_and_price else self.filled_quantity
        # consider worse case taker fees when using use_origin_quantity_and_price as the order is not filled yet
        taker_or_maker = enums.ExchangeConstantsOrderColumns.TAKER.value \
            if use_origin_quantity_and_price else self.taker_or_maker
        if self.is_cleared():
            # order is cleared, it might have been filled or cancelled. Use existing fees
            return copy.copy(self.fee)
        if self.fees_currency_side is enums.FeesCurrencySide.UNDEFINED:
            computed_fee = self.exchange_manager.exchange.get_trade_fee(
                self.symbol, self.order_type, quantity, price, taker_or_maker
            )
            value = computed_fee[enums.FeePropertyColumns.COST.value]
            currency = computed_fee[enums.FeePropertyColumns.CURRENCY.value]
            is_from_exchange = computed_fee[enums.FeePropertyColumns.IS_FROM_EXCHANGE.value]
        else:
            symbol_fees = self.exchange_manager.exchange.get_fees(self.symbol)
            fees = decimal.Decimal(f"{symbol_fees[taker_or_maker]}")
            if self.fees_currency_side is enums.FeesCurrencySide.CURRENCY:
                value = quantity / price * fees
                currency = self.currency
            else:
                value = quantity * price * fees
                currency = self.market
        return {
            enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: is_from_exchange,
            enums.FeePropertyColumns.COST.value: forced_value if forced_value is not None else value,
            enums.FeePropertyColumns.CURRENCY.value: currency,
        }

    def get_profitability(self):
        if self.filled_price != 0 and self.created_last_price != 0:
            if self.filled_price >= self.created_last_price:
                self.order_profitability = 1 - self.filled_price / self.created_last_price
                if self.side is enums.TradeOrderSide.SELL:
                    self.order_profitability *= -1
            else:
                self.order_profitability = 1 - self.created_last_price / self.filled_price
                if self.side is enums.TradeOrderSide.BUY:
                    self.order_profitability *= -1
        return self.order_profitability

    async def default_exchange_update_order_status(self):
        raw_order = await self.exchange_manager.exchange.get_order(
            self.exchange_order_id, self.symbol, order_type=self.order_type
        )
        new_status = order_util.parse_order_status(raw_order)
        self.is_synchronized_with_exchange = True
        if new_status in {enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED}:
            await self.on_fill()
        elif new_status is enums.OrderStatus.CANCELED:
            await self.trader.cancel_order(self)

    def generate_executed_time(self):
        return self.exchange_manager.exchange.get_exchange_current_time()

    def is_counted_in_available_funds(self) -> bool:
        if self.state and self.state.is_already_counted_in_available_funds:
            return False
        if self.is_active:
            if self.trader.simulate or (
                self.reduce_only and self.exchange_manager and self.exchange_manager.is_future
            ):
                return not (
                    # in trading simulator, stop orders and TP do not lock funds
                    # when trading futures, reduce only stop loss and take profit orders do not lock funds
                    order_util.is_stop_order(self.order_type) or order_util.is_take_profit_order(self.order_type)
                )
            # lock funds when not self-managed
            return not self.is_self_managed()
        # inactive: not locking funds
        return False

    def is_self_managed(self):
        if self.is_cleared():
            return order_util.is_stop_order(self.order_type) or order_util.is_take_profit_order(self.order_type)
        return self.trader.allow_artificial_orders and \
            not self.is_synchronized_with_exchange and \
            not self.exchange_manager.exchange.is_supported_order_type(self.order_type)

    def is_long(self):
        return self.side is enums.TradeOrderSide.BUY

    def is_short(self):
        return self.side is enums.TradeOrderSide.SELL

    def is_partially_filled(self) -> bool:
        return constants.ZERO < self.filled_quantity < self.origin_quantity

    def get_remaining_quantity(self) -> decimal.Decimal:
        return self.origin_quantity - self.filled_quantity

    def get_locked_quantity(self) -> decimal.Decimal:
        return self.filled_quantity if (
            self.is_filled() or self.USE_ORIGIN_QUANTITY_AS_FILLED_QUANTITY
        ) else self.get_remaining_quantity()

    def get_remaining_cost(self) -> decimal.Decimal:
        return self.get_cost(self.get_remaining_quantity())

    def update_from_raw(self, raw_order):
        if self.side is None or self.order_type is None:
            try:
                self._update_type_from_raw(raw_order)
                if self.taker_or_maker is None:
                    self._update_taker_maker()
            except KeyError:
                logging.get_logger(self.__class__.__name__).warning("Failed to parse order side and type")

        # use stop price when available
        price = (
            raw_order.get(enums.ExchangeConstantsOrderColumns.PRICE.value, None)
            or raw_order.get(enums.ExchangeConstantsOrderColumns.STOP_PRICE.value, None)
            or raw_order.get(enums.ExchangeConstantsOrderColumns.STOP_LOSS_PRICE.value, None)
            or raw_order.get(enums.ExchangeConstantsOrderColumns.TAKE_PROFIT_PRICE.value, None)
            or 0.0
        )
        filled_price = decimal.Decimal(str(price))
        # set average price with real average price if available, use filled_price otherwise
        average_price = decimal.Decimal(str(
            raw_order.get(enums.ExchangeConstantsOrderColumns.AVERAGE.value, 0.0) or filled_price
        ))

        return self.update(
            symbol=str(raw_order.get(enums.ExchangeConstantsOrderColumns.SYMBOL.value, None)),
            current_price=decimal.Decimal(str(price)),
            quantity=decimal.Decimal(str(raw_order.get(enums.ExchangeConstantsOrderColumns.AMOUNT.value, 0.0) or 0.0)),
            price=decimal.Decimal(str(price)),
            status=order_util.parse_order_status(raw_order),
            order_id=raw_order.get(enums.ExchangeConstantsOrderColumns.ID.value, None),
            exchange_order_id=str(raw_order.get(enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value, None)),
            quantity_filled=decimal.Decimal(str(raw_order.get(enums.ExchangeConstantsOrderColumns.FILLED.value, 0.0)
                                                or 0.0)),
            filled_price=decimal.Decimal(str(filled_price)),
            average_price=decimal.Decimal(str(average_price)),
            total_cost=decimal.Decimal(str(raw_order.get(enums.ExchangeConstantsOrderColumns.COST.value, 0.0) or 0.0)),
            fee=order_util.parse_raw_fees(raw_order.get(enums.ExchangeConstantsOrderColumns.FEE.value, None)),
            timestamp=raw_order.get(enums.ExchangeConstantsOrderColumns.TIMESTAMP.value, None),
            reduce_only=raw_order.get(enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value, False),
            trigger_above=raw_order.get(enums.ExchangeConstantsOrderColumns.TRIGGER_ABOVE.value, None)
        )

    async def update_from_order(self, other_order):
        self.is_synchronized_with_exchange = other_order.is_synchronized_with_exchange
        self.is_from_this_octobot = other_order.is_from_this_octobot

        self.origin_quantity = other_order.origin_quantity
        self.origin_price = other_order.origin_price
        self.origin_stop_price = other_order.origin_stop_price
        self.symbol = other_order.symbol
        self.currency = other_order.currency
        self.market = other_order.market
        self.quantity_currency = other_order.quantity_currency
        self.taker_or_maker = other_order.taker_or_maker
        self.side = other_order.side

        self.order_id = other_order.order_id
        self.exchange_order_id = other_order.exchange_order_id
        self.status = other_order.status

        self.filled_quantity = other_order.filled_quantity
        self.filled_price = other_order.filled_price
        self.fee = other_order.fee
        self.fees_currency_side = other_order.fees_currency_side
        self.total_cost = other_order.total_cost
        self.order_profitability = other_order.order_profitability
        self.executed_time = other_order.executed_time

        self.canceled_time = other_order.canceled_time

        self.reduce_only = other_order.reduce_only

        self.position_side = other_order.position_side

        self.is_waiting_for_chained_trigger = other_order.is_waiting_for_chained_trigger

        if other_order.state is not None:
            await other_order.state.replace_order(self)

    def update_from_storage_order_details(self, order_details):
        # rebind order attributes that are not stored on exchange
        order_dict = order_details.get(constants.STORAGE_ORIGIN_VALUE, {})
        self.tag = order_dict.get(enums.ExchangeConstantsOrderColumns.TAG.value, self.tag)
        self.broker_applied = order_dict.get(
            enums.ExchangeConstantsOrderColumns.BROKER_APPLIED.value,
            self.broker_applied
        )
        self.order_id = order_dict.get(enums.ExchangeConstantsOrderColumns.ID.value, self.order_id)
        self.exchange_order_id = order_dict.get(enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value,
                                                self.exchange_order_id)
        self.taker_or_maker = enums.ExchangeConstantsMarketPropertyColumns(
            order_dict[enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value]
        ).value if order_dict.get(enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value) else self.taker_or_maker
        self.is_active = order_dict.get(enums.ExchangeConstantsOrderColumns.IS_ACTIVE.value, self.is_active)
        if active_trigger := order_details.get(enums.StoredOrdersAttr.ACTIVE_TRIGGER.value):
            active_trigger_price = (
                decimal.Decimal(str(active_trigger[enums.StoredOrdersAttr.ACTIVE_TRIGGER_PRICE.value]))
                if active_trigger.get(enums.StoredOrdersAttr.ACTIVE_TRIGGER_PRICE.value) else None
            )
            active_trigger_above = active_trigger.get(enums.StoredOrdersAttr.ACTIVE_TRIGGER_ABOVE.value)
            if active_trigger_price is not None and active_trigger_above is not None:
                self.use_active_trigger(
                    order_util.create_order_price_trigger(self, active_trigger_price, active_trigger_above)
                )
            else:
                logging.get_logger(self.__class__.__name__).error(
                    f"Ignored unknown trigger configuration: {active_trigger}"
                )
        if cancel_policy := order_details.get(enums.StoredOrdersAttr.CANCEL_POLICY.value):
            self.cancel_policy = cancel_policy_factory.create_cancel_policy(
                cancel_policy[enums.StoredOrdersAttr.CANCEL_POLICY.value],
                cancel_policy[enums.StoredOrdersAttr.CANCEL_KWARGS.value],
            )
        self.trader_creation_kwargs = order_details.get(enums.StoredOrdersAttr.TRADER_CREATION_KWARGS.value,
                                                        self.trader_creation_kwargs)
        self.exchange_creation_params = order_details.get(enums.StoredOrdersAttr.EXCHANGE_CREATION_PARAMS.value,
                                                          self.exchange_creation_params)
        self.has_been_bundled = order_details.get(enums.StoredOrdersAttr.HAS_BEEN_BUNDLED.value,
                                                  self.has_been_bundled)
        self.associated_entry_ids = order_details.get(enums.StoredOrdersAttr.ENTRIES.value,
                                                      self.associated_entry_ids)
        self.update_with_triggering_order_fees = order_details.get(
            enums.StoredOrdersAttr.UPDATE_WITH_TRIGGERING_ORDER_FEES.value, False
        )
        if trailing_profile := order_details.get(enums.StoredOrdersAttr.TRAILING_PROFILE.value):
            self.trailing_profile = trailing_profiles.create_trailing_profile(
                trailing_profiles.TrailingProfileTypes(
                    trailing_profile[enums.StoredOrdersAttr.TRAILING_PROFILE_TYPE.value]
                ),
                trailing_profile[enums.StoredOrdersAttr.TRAILING_PROFILE_DETAILS.value],
            )

    def consider_as_filled(self):
        self.status = enums.OrderStatus.FILLED
        if self.executed_time == 0:
            self.executed_time = self.timestamp
        if self.filled_quantity == constants.ZERO:
            self.filled_quantity = self.origin_quantity
        if self.filled_price == constants.ZERO:
            self.filled_price = self.origin_price
        self._update_total_cost()

    def _update_total_cost(self):
        # use filled amounts when available
        quantity = self.filled_quantity if self.filled_quantity else self.origin_quantity
        self.total_cost = self.get_cost(quantity)

    def get_cost(self, quantity: decimal.Decimal) -> decimal.Decimal:
        price = self.filled_price if self.filled_price else self.origin_price
        if self.quantity_currency == self.currency:
            # quantity in BTC for BTC/USDT => cost = BTC * price(BTC in USDT)
            return quantity * price
        else:
            # quantity in USDT for BTC/USDT => cost = price(BTC in USDT)
            return quantity

    def update_order_filled_values(self, ideal_price: decimal.Decimal):
        if not self.filled_price:
            # keep order.filled_price if already set (!= 0)
            self.filled_price = order_util.get_valid_filled_price(self, ideal_price)
        if not self.filled_quantity or self.exchange_manager.trader.simulate:
            # keep self.filled_quantity if already set (!= 0) in real trading
            self.filled_quantity = self.origin_quantity
        self._update_total_cost()

    def consider_as_canceled(self):
        self.status = enums.OrderStatus.CANCELED
        if self.canceled_time == 0:
            self.canceled_time = self.timestamp

    def update_order_from_raw(self, raw_order):
        self.status = order_util.parse_order_status(raw_order)
        self.total_cost = decimal.Decimal(str(raw_order[enums.ExchangeConstantsOrderColumns.COST.value] or 0))
        self.filled_quantity = decimal.Decimal(str(raw_order[enums.ExchangeConstantsOrderColumns.FILLED.value] or 0))
        self.filled_price = decimal.Decimal(str(raw_order[enums.ExchangeConstantsOrderColumns.PRICE.value] or 0))
        if not self.filled_price and self.filled_quantity:
            self.filled_price = self.total_cost / self.filled_quantity

        self._update_taker_maker()

        self.fee = order_util.parse_raw_fees(raw_order[enums.ExchangeConstantsOrderColumns.FEE.value])

        self.executed_time = self.trader.exchange_manager.exchange.get_uniformized_timestamp(
            raw_order[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value])

    def _update_type_from_raw(self, raw_order):
        try:
            self.exchange_order_type = enums.TradeOrderType(raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value])
        except ValueError:
            self.exchange_order_type = enums.TradeOrderType.UNKNOWN
        self.side, self.order_type = parse_order_type(raw_order)

    def _get_open_price(self):
        if self.triggered_by is None:
            return self.created_last_price
        return self.triggered_by.get_filling_price()  # pylint: disable=protected-access

    def get_filling_price(self):
        return self.origin_price

    def _should_instant_fill(self):
        # default implementation, to be overridden in subclasses
        return False

    def _update_taker_maker(self):
        if self.order_type in [enums.TraderOrderType.SELL_MARKET,
                               enums.TraderOrderType.BUY_MARKET,
                               enums.TraderOrderType.STOP_LOSS]:
            # always true
            self.taker_or_maker = enums.ExchangeConstantsMarketPropertyColumns.TAKER.value
        else:
            # true 90% of the time: impossible to know for sure the reality
            self.taker_or_maker = (
                enums.ExchangeConstantsMarketPropertyColumns.TAKER if self._should_instant_fill()
                else enums.ExchangeConstantsMarketPropertyColumns.MAKER
            ).value

    def is_synchronization_enabled(self):
        return (
            self.exchange_manager is not None and
            self.exchange_manager.exchange_personal_data.orders_manager.enable_order_auto_synchronization
        )

    def to_dict(self):
        filled_price = self.filled_price if self.filled_price > 0 else self.origin_price
        return {
            enums.ExchangeConstantsOrderColumns.ID.value: self.order_id,
            enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value: self.exchange_order_id,
            enums.ExchangeConstantsOrderColumns.SYMBOL.value: self.symbol,
            enums.ExchangeConstantsOrderColumns.PRICE.value: filled_price,
            enums.ExchangeConstantsOrderColumns.STATUS.value: self.status.value,
            enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: self.creation_time or self.timestamp,
            enums.ExchangeConstantsOrderColumns.TYPE.value: self.exchange_order_type.value
            if self.exchange_order_type else None,
            enums.ExchangeConstantsOrderColumns.SIDE.value: self.side.value,
            enums.ExchangeConstantsOrderColumns.AMOUNT.value: self.origin_quantity,
            enums.ExchangeConstantsOrderColumns.COST.value: self.total_cost,
            enums.ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value: self.quantity_currency,
            enums.ExchangeConstantsOrderColumns.FILLED.value: self.filled_quantity,
            enums.ExchangeConstantsOrderColumns.FEE.value: self.fee,
            enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value: self.reduce_only,
            enums.ExchangeConstantsOrderColumns.TAG.value: self.tag,
            enums.ExchangeConstantsOrderColumns.TRIGGER_ABOVE.value: self.trigger_above,
            enums.ExchangeConstantsOrderColumns.SELF_MANAGED.value: self.is_self_managed(),
            enums.ExchangeConstantsOrderColumns.BROKER_APPLIED.value: self.broker_applied,
            enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value: self.taker_or_maker,
            enums.ExchangeConstantsOrderColumns.IS_ACTIVE.value: self.is_active,
        }

    def clear_active_order_elements(self):
        if self.state is not None:
            self.state.clear()


    def clear(self):
        if self.active_trigger:
            self.active_trigger.clear()
        self.clear_active_order_elements()
        self.trader = None # type: ignore
        self.exchange_manager = None # type: ignore
        self.trader_creation_kwargs = {}

    def is_cleared(self):
        return self.exchange_manager is None

    def to_string(self):
        inactive = "" if self.is_active else "[Inactive] "
        chained_order = "" if self.triggered_by is None else \
            "triggered chained order | " if self.is_created() else "untriggered chained order | "
        tag = f" | tag: {self.tag}" if self.tag else ""
        fees = (
            f"Fees : {self.fee[enums.FeePropertyColumns.COST.value]} {self.fee[enums.FeePropertyColumns.CURRENCY.value]} | "
            if self.fee else ""
        )
        trailing_profile = f"Trailing profile : {self.trailing_profile} | " if self.trailing_profile else ""
        cancel_policy = f"Cancel policy : {self.cancel_policy} | " if self.cancel_policy else ""
        filled_quantity = f" ({self.filled_quantity} Filled)" if self.filled_quantity else ""
        return (
            f"{inactive}{self.symbol} | "
            f"{chained_order}"
            f"{self.order_type.name if self.order_type is not None else 'Unknown'} | "
            f"Price : {str(self.origin_price)} | "
            f"Quantity : {str(self.origin_quantity)}{filled_quantity}{' (Reduce only)' if self.reduce_only else ''} | "
            f"State : {self.state.state.value if self.state is not None else 'Unknown'} | "
            f"{trailing_profile}"
            f"{cancel_policy}"
            f"{fees}"
            f"id : {self.order_id}{tag} "
            f"exchange id: {self.exchange_order_id}"
        )

    def __str__(self):
        return self.to_string()

    def is_to_be_maintained(self):
        return self.trader is not None


def parse_order_type(raw_order) -> (enums.TradeOrderSide, enums.TraderOrderType):
    try:
        side: enums.TradeOrderSide = enums.TradeOrderSide(raw_order[enums.ExchangeConstantsOrderColumns.SIDE.value])
        order_type: enums.TradeOrderType = enums.TradeOrderType.UNKNOWN
        parsed_order_type: typing.Optional[enums.TraderOrderType] = enums.TraderOrderType.UNKNOWN
        try:
            order_type = enums.TradeOrderType(raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value])
        except ValueError as e:
            if raw_order[enums.ExchangeConstantsOrderColumns.TYPE.value] is None:
                # Last chance: try to infer order type from taker / maker status
                if enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value in raw_order:
                    return side, _infer_order_type_from_maker_or_taker(raw_order, side)
                # No order type info: use unknown order type
                return side, enums.TraderOrderType.UNKNOWN
            else:
                # Incompatible order type info: raise error
                raise e

        if order_type is enums.TradeOrderType.UNKNOWN:
            parsed_order_type = enums.TraderOrderType.UNKNOWN
        elif order_type is enums.TradeOrderType.UNSUPPORTED:
            parsed_order_type = enums.TraderOrderType.UNSUPPORTED
        elif side is enums.TradeOrderSide.BUY:
            if order_type is enums.TradeOrderType.LIMIT or order_type == enums.TradeOrderType.LIMIT_MAKER:
                parsed_order_type = enums.TraderOrderType.BUY_LIMIT
            elif order_type is enums.TradeOrderType.MARKET:
                parsed_order_type = enums.TraderOrderType.BUY_MARKET
            else:
                parsed_order_type = _get_sell_and_buy_types(order_type)
        elif side is enums.TradeOrderSide.SELL:
            if order_type is enums.TradeOrderType.LIMIT or order_type is enums.TradeOrderType.LIMIT_MAKER:
                parsed_order_type = enums.TraderOrderType.SELL_LIMIT
            elif order_type is enums.TradeOrderType.MARKET:
                parsed_order_type = enums.TraderOrderType.SELL_MARKET
            else:
                parsed_order_type = _get_sell_and_buy_types(order_type)
        return side, parsed_order_type
    except (KeyError, ValueError):
        return None, None


def _infer_order_type_from_maker_or_taker(raw_order, side):
    is_taker = raw_order[enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value] \
               == enums.ExchangeConstantsOrderColumns.TAKER.value
    if side is enums.TradeOrderSide.BUY:
        if is_taker:
            return enums.TraderOrderType.BUY_MARKET
        return enums.TraderOrderType.BUY_LIMIT
    if is_taker:
        return enums.TraderOrderType.SELL_MARKET
    return enums.TraderOrderType.SELL_LIMIT


def _get_trade_order_type(order_type: enums.TraderOrderType) -> typing.Optional[enums.TradeOrderType]:
    if order_type is enums.TraderOrderType.BUY_LIMIT or order_type is enums.TraderOrderType.SELL_LIMIT:
        return enums.TradeOrderType.LIMIT
    if order_type is enums.TraderOrderType.BUY_MARKET or order_type is enums.TraderOrderType.SELL_MARKET:
        return enums.TradeOrderType.MARKET
    elif order_type is enums.TraderOrderType.TAKE_PROFIT:
        return enums.TradeOrderType.TAKE_PROFIT
    elif order_type is enums.TraderOrderType.TAKE_PROFIT_LIMIT:
        return enums.TradeOrderType.TAKE_PROFIT_LIMIT
    elif order_type is enums.TraderOrderType.STOP_LOSS:
        return enums.TradeOrderType.STOP_LOSS
    elif order_type is enums.TraderOrderType.STOP_LOSS_LIMIT:
        return enums.TradeOrderType.STOP_LOSS_LIMIT
    elif order_type is enums.TraderOrderType.TRAILING_STOP:
        return enums.TradeOrderType.TRAILING_STOP
    elif order_type is enums.TraderOrderType.TRAILING_STOP_LIMIT:
        return enums.TradeOrderType.TRAILING_STOP_LIMIT
    return None


def _get_sell_and_buy_types(order_type) -> typing.Optional[enums.TraderOrderType]:
    if order_type is enums.TradeOrderType.STOP_LOSS:
        return enums.TraderOrderType.STOP_LOSS
    elif order_type is enums.TradeOrderType.STOP_LOSS_LIMIT:
        return enums.TraderOrderType.STOP_LOSS_LIMIT
    elif order_type is enums.TradeOrderType.TAKE_PROFIT:
        return enums.TraderOrderType.TAKE_PROFIT
    elif order_type is enums.TradeOrderType.TAKE_PROFIT_LIMIT:
        return enums.TraderOrderType.TAKE_PROFIT_LIMIT
    elif order_type is enums.TradeOrderType.TRAILING_STOP:
        return enums.TraderOrderType.TRAILING_STOP
    elif order_type is enums.TradeOrderType.TRAILING_STOP_LIMIT:
        return enums.TraderOrderType.TRAILING_STOP_LIMIT
    return None
